#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""Data Representation module

This module contains the core classes and function for manipulating test data
in ttproto. This moudule provides commodity functions for:
 - generating messages
 - matching a received message with some expected pattern
 - encoding/decoding messages

All data manipulated for these purposes derive from the root class Data. There
are two variants of data: values and templates. Values are used to represent an
actual value (eg. an integer, an IPv6 packet, ...) and templates are used to
represent a pattern of values (eg. Range(int, 0, 15) means “any integer between
0 and 15”).

The figure below is a partial representation of the hierarchy of classes
deriving from Data.

                                           .------.
                            .------------->| Data |<-------.
                            |              '------'        |
                            |                              |
                        .-------.                    .----------.
           .----------->| Value |<-----.             | Template |<----.
           |            '-------'      |             '----------'     |
           |                           |                ^     ^       |
           |                           |                |     |       |
           |                           |                |     |       |
  .----------------.            .-------------.    .-------.  |       |
  | PrimitiveValue |            | PacketValue |    | Range |  |       |
  '----------------'            '-------------'    '-------'  |       |
    ^              ^                ^       ^           .-----------. |
    | .-------.    | .-----.        |       |           | ValueList | |
    | | bytes |    | | int |    .------. .-----.        '-----------' |
    | '-------'    | '-----'    | IPv6 | | UDP |                      |
    |    ^         |    ^       '------' '-----'               .-------------.
    |    |         |    |                                      | IPv6Prefix  |
 .------------. .----------.                                   '-------------'
 | BytesValue | | IntValue |
 '------------' '----------'
        ^
        |
 .-------------.
 | IPv6Address |
 '-------------'


Primitive values are handled in python with a set of predefined classes (int,
str, bytes, float, ...). For consistency purpose, ttproto provides a subclass for
each of these types that derives on Value (IntValue, StrValue, BytesValue). Most
classes and functions will perform the conversion on the fly using as_data() or
store_data(), thus it is not necessary to instantiate instances of these classes
directly. For example the following instances of the Range class will be
equivalent:
    Range (int, 0, 15)
    Range (IntValue, IntValue(0), IntValue(15))

Data objects are immutable (FIXME: to be confirmed). Once they are created,
their content cannot be modified.

Each Data object is associted to a type, which in the terminology of ttproto is a
subclass of Value. For an instance of a class derived from Value, get_type()
will return type(obj). An instance of a template will return the type used at
initialisation time. For example: Range(int, 0, 15).get_type() == IntValue.

Structured values (eg: PacketValue, ListValue) will store a separate Data object
for each of their fields. These fields can contain templates too.

Each Data object can be derived from another Data object. The resulting object
will be a combination of the parent object and of the child object. Deriving an
object is done by calling the function operator ( ).

For example:
    # an IPv6 header sent by 2001:db8::1
    #
    IPv6_src_A = IPv6 (src="2001:db8::1")

    # an IPv6 header multicasted by A
    #
    # (note that IPv6Prefix is a template for the type IPv6Address
    #  this one match all addresses starting with "ff")
    IPv6_multicast_from_A = IPv6_src_A (dst = IPv6Prefix ("ff00::/8"))

    # an IPv6 header sent by A to all nodes on the link
    #
    IPv6_A_to_all_nodes = IPv6_multicast_from_A (dst = "ff02::1")


This example will create a hierarchy of IPv6 value instances. Each of them
adding some information about what we want to have in our IPv6 header.

                                  .------------------------------------.
                                  |             IPv6_src_A             |
                                  | type: IPv6                         |
                                  |                                    |
                                  | fields: {                          |
                                  |   src: IPv6Address ("2001:db8::1") |
                                  | }                                  |
                                  '------------------------------------'
                     .------------------------------------.  ^
                     |       IPv6_multicast_from_A        | /
                     | type:   IPv6                       |/
                     | parent: IPv6_src_A                 |
                     |                                    |
                     | fields: {                          |
                     |   dst: IPv6Prefix ("ff::/8")       |
                     | }                                  |
                     '------------------------------------'
.------------------------------------.  ^
|        IPv6_A_to_all_nodes         | /
| type:   IPv6                       |/
| parent: IPv6_multicast_from_A      |
|                                    |
| fields: {                          |
|   dst: IPv6Address ("ff02::1")     |
| }                                  |
'------------------------------------'

Now these data can be used for several purposes, especially:
 - matching incoming messages
 - generating output messages

In case we want to match an incoming message from A sent to the all-nodes
multicast address, we can use the value IPv6_A_to_all_nodes. The match() method
will check that the given message matches all of the fields of
IPv6_A_to_all_nodes and of its parent values (IPv6_multicast_from_A and
IPv6_src_A).

Examples:
    >>> IPv6_A_to_all_nodes.match(IPv6(src="2001:db8::1", dst="ff02::1"))
    True
    >>> IPv6_A_to_all_nodes.match(IPv6(src="2001:db8::2", dst="ff02::1"))
    False
    >>> IPv6_A_to_all_nodes.match(IPv6(src="2001:db8::1", dst="ff02::2"))
    False

The second pattern IPv6_multicast_from_A may seem superfluous here because it
the third pattern IPv6_A_to_all_nodes is more strict. However it will be useful
because ttproto will be able to report which pattern in the hierarchy did or did not
match. For example:
    Match result
    - IPv6_from_A        -> success
    - IPv6_multicast_from_A    -> success
    - IPv6_A_to_all_nodes    -> failure
      * DestinationAddress: ValueMismatch: expected ff02::1 (got ff02::5)

Thus, with a quick look at the name of the patterns, it is easy to learn that it
is a multicast message sent by A but that it was not sent to all nodes. This
information will be useful later to interpret the test results in case of a
fail: with a quick look it is possible to know what did not match and whether it
is related to the test purpose or not.

It is possible to generate a data whose content is incompatible with its
parents. For example:
    >>> IPv6_multicast_from_A(dst="2001:db8::2")
    IPv6(dst='2001:db8::2').set_parent(IPv6(dst=IPv6Prefix('ff00::/8')).set_parent(IPv6(src='2001:db8::1')))

However such a data will not be useful in practice. If it is used for matching
another data, then the match will always fail (because the same destination
address cannot match 2001:db8::2 and ff00::/8). It will not be possible to
flatten the value (see below) and generate a message from this data.


Flat Values
-----------

It is possible to mix values and templates and to have multiple values for the
same field (because of inheritance). But in order to generate and exchange real
messages, we must have a unique unambiguous value. This kind of value are called
« Flat Values ». To be considered flat, a values fulfill the following
conditions:
    - it must not contain any template
    - it must not be derived from another data
    - in case of structured values, all its fields must either contain flat
      values or be undefined

Values can be collapsed into a flat value by calling the flatten() method. It
will return a new Value object with the same content, but flattened:

    >>> IPv6_A_to_all_nodes
    IPv6(dst='ff02::1').set_parent(IPv6(dst=IPv6Prefix('ff00::/8')).set_parent(IPv6(src='2001:db8::1')))

    >>> IPv6_A_to_all_nodes.flatten()
    IPv6(src='2001:db8::1', dst='ff02::1')

All values cannot be flattened. The operation will fail if a value does not
match other values or templates set for the same field.

    >>> IPv6_multicast_from_A(dst="2001:db8::2").flatten()
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "/home/abaire/git/ttproto/core/data.py", line 656, in flatten
        result = value_class._flatten (values)
      File "/home/abaire/git/ttproto/core/packet.py", line 547, in _flatten
        new_v = Data.flatten (*candidates)
      File "/home/abaire/git/ttproto/core/data.py", line 661, in flatten
        raise exceptions.Error ("flattened value does not comply with one of its templates")
    core.exceptions.Error: Error: flattened value does not comply with one of its templates


Messages
--------

Messages are the basic unit exchanged between entities in the test. A message
has a value representation (which is flat and fully defined) and a binary
representation.

There are two ways to build a message:
- from a value : in this case, the constructor of Message will flatten the value
  and call the method build_message() to generate the message. This functions
  perform several interesting operations:
   * it fills undefined fields with a default value or for some special cases
   (checksum, type, length) it computes a value on the fly
   * it encodes the message into its binary format

    >>> Message(UInt16(66)).display()
    ###[ UInt16 ]###
      Value=                    66
    Encoded as:
        00 42

    >>> Message(IPv6(src="2001:db8:1::1", dst="2001:db8:2::1")/UDP(sport=1025, dport=427)/"blah blah").display()
    ###[ IPv6 ]###
      Version=                  6
      TrafficClass=             0
      FlowLabel=                0
      PayloadLength=            17
      NextHeader=               17
      HopLimit=                 64
      SourceAddress=            2001:db8:1::1
      DestinationAddress=       2001:db8:2::1
      Payload=
    ###[ UDP ]###
        SourcePort=             1025
        DestinationPort=        427
        Length=                 17
        Checksum=               58896
        Payload=
    ###[ StrValue ]###
          Value=                'blah blah'
    Encoded as:
        60 00 00 00 00 11 11 40  20 01 0d b8 00 01 00 00
        00 00 00 00 00 00 00 01  20 01 0d b8 00 02 00 00
        00 00 00 00 00 00 00 01  04 01 01 ab 00 11 e6 10
        62 6c 61 68 20 62 6c 61  68

- from a binary string and a type : in this case, the constructor of Message
  will call decode_message() on the expected type to decode the binary string
  and generate the associated value

    >>> Message (b"\0\x42", UInt16).display()
    ###[ UInt16 ]###
      Value=                    66
    Encoded as:
        00 42

    >>> Message(b'`\x00\x00\x00\x00\x11\x11@ \x01\r\xb8\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01 \x01\r\xb8\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x04\x01\x01\xab\x00\x11\xe6\x10blah blah', IPv6).display()
    ###[ IPv6 ]###
      Version=                  6
      TrafficClass=             0
      FlowLabel=                0
      PayloadLength=            17
      NextHeader=               17
      HopLimit=                 64
      SourceAddress=            2001:db8:1::1
      DestinationAddress=       2001:db8:2::1
      Payload=
    ###[ UDP ]###
        SourcePort=             1025
        DestinationPort=        427
        Length=                 17
        Checksum=               58896
        Payload=
    ###[ BytesValue ]###
          Value=                b'blah blah'
    Encoded as:
        60 00 00 00 00 11 11 40  20 01 0d b8 00 01 00 00
        00 00 00 00 00 00 00 01  20 01 0d b8 00 02 00 00
        00 00 00 00 00 00 00 01  04 01 01 ab 00 11 e6 10
        62 6c 61 68 20 62 6c 61  68

"""

from contextlib import contextmanager
import re, sys

from ttproto.core.typecheck import typecheck, optional, iterable, this_class, either, with_attr
from ttproto.core import named, exceptions

__all__ = [
    'is_type',
    'is_data',
    'is_value',
    'is_flat_value',
    'is_binary',
    'get_binary_length',
    'is_flatvalue_binslice',
    'is_flatvalue_binary',
    'concatenate',
    'get_type',
    'BinarySlice',
    'Data',
    'as_data',
    'store_data',
    'pack',
    'Value',
    'Omit',
    'Template',
    'Mismatch',
    'ValueMismatch',
    'TypeMismatch',
    'VariantMismatch',
    'LengthMismatch',
    'TemplateMismatch',
    'Message',
    'MessageDescription',
    'Bidict',
    'BidictValueType',
    'DifferenceList',
]

__known_primitive_types = int, str, bytes, bool


def is_type(arg):
    """Return True if the argument is a subclass of Value or if it is one of
    the known primitive types: int, str, bytes and bool (the types that can
    be mapped to one class deriving from PrimitiveValue)
    """
    return (isinstance(arg, type) and (issubclass(arg, Value))) or arg in __known_primitive_types


def is_data(arg):
    """Return True if the argument is an instance of the Data class or of
    one of the known primitive types (see is_type())
    """
    return isinstance(arg, Data) or type(arg) in __known_primitive_types


def is_value(arg):
    """Return True if the argument is an instance of the Value class or of
    one of the known primitive types (see is_type())
    """
    return isinstance(arg, Value) or type(arg) in __known_primitive_types


def is_flat_value(arg):
    """Return True if the argument is an instance of the Value class or of
    one of the known primitive types (see is_type()) and if it is flat
    """
    if isinstance(arg, Value):
        return arg.is_flat()
    else:
        return type(arg) in (int, str, bytes)


def is_binary(arg):
    """Return True is the argement is a binary object, which can be one of
    the following:
     - an instance of bytes
     - a tuple in the form (bytes, int)
    The second format is used to represent bit strings that do not end at a
    byte boundary. The integer number represents the number of bits in the
    last byte to be included ; its value can range between 0 and 7. The bits
    included in the last byte are the most significant ones.

    Examples:
        the 3-bit string '011' is mapped as: (b"\x60", 3)
        the 10-bit string '1111000011' is mapped as: (b"\xf0\xc0", 2)
        the 16-bit string '0000000101111111' is mapped as b"\x01\x7f"
    """
    return isinstance(arg, bytes) or (type(arg) == tuple and
                                      len(arg) == 2 and
                                      isinstance(arg[0], bytes) and
                                      type(arg[1]) == int and
                                      0 <= arg[1] <= 7)


@typecheck
def get_binary_length(arg: is_binary):
    """Return the length in bits of a binary object"""
    if isinstance(arg, bytes):
        return len(arg) * 8
    else:
        return len(arg[0]) * 8 + arg[1] - 8


def is_flatvalue_binslice(arg):
    """Return True if the argument is a tuple containing a flat Value object
    and a BinarySlice

    (format of the result of Value.decode_message())
    """

    return (type(arg) == tuple
            and len(arg) == 2
            and isinstance(arg[0], Value)
            and is_flat_value(arg[0])
            and type(arg[1]) == BinarySlice)


def is_flatvalue_binary(arg):
    """Return True if the argument is a tuple containing a flat Value object
    and a binary object (is_binary())

    (format of the result of Value.build_message())
    """
    return (type(arg) == tuple
            and len(arg) == 2
            and isinstance(arg[0], Value) and arg[0].is_flat()
            and is_binary(arg[1]))


@typecheck
def concatenate(k: iterable) -> is_binary:
    """concatenate multible binary objects (is_binary()) together

    Limitations: the current implementation accepts any binary objects as
    input, however the result of the function must end at a byte boundary
    """
    # TODO: use the 'struct' package to optimise this
    bins = []
    buff = ""
    for b in k:
        assert is_binary(b)
        #        print "*", repr(b)
        if isinstance(b, bytes) and not buff:
            # byte aligned -> nothing to do
            #            print repr(b)
            bins.append(b)
        else:
            for c in (b if isinstance(b, bytes) else b[0]):
                buff += format(c, "08b")
            if type(b) == tuple:
                buff = buff[:-8 + b[1]]
            # print "buff:", buff, repr (b)

            if len(buff) % 8 == 0:
                bins.append(bytes([eval("0b" + buff[i:i + 8]) for i in range(0, len(buff), 8)]))
                #                    print "->", repr(bins[-1])
                buff = ""

    assert not buff  # TODO: support non 8-bit aligned data ???

    return b"".join(bins)


@typecheck
def get_type(arg: either(is_data, is_type)) -> type:
    """Return the type (subclass of Value) associated to the argument.

     - if arg is a Value class     -> return arg
     - if arg is a Value instance     -> return type (arg)
     - if arg is a Template instance -> return the arg.get_type()
     - if arg is a primitive type or a primitive value ->    return the corresponding PrimitiveValue class

    The result will always be a subclass of Value

    Examples:
        >>> get_type(IntValue(45))
        <class 'core.primitive.IntValue'>

        >>> get_type(45)
        <class 'core.primitive.IntValue'>

        >>> get_type(int)
        <class 'core.primitive.IntValue'>

        >>> get_type(IntValue)
        <class 'core.primitive.IntValue'>

        >>> get_type(Range(int, 0, 15))
        <class 'core.primitive.IntValue'>

        >>> get_type(IPv6 (nh=45, len=48))
        <class 'core.lib.inet.ipv6.IPv6'>
    """

    if isinstance(arg, type):
        if issubclass(arg, Value):
            # Value class
            return arg
    else:
        if isinstance(arg, Data):
            # Data instance
            return arg.get_type()

        arg = type(arg)

    # assume that it is a primitive type
    assert arg in __known_primitive_types

    return primitive.PrimitiveValue.get_type_for_python_type(arg)


class BinarySlice:
    """Represent a slice of binary data that can begin or end at any bit
    (not necessary at a byte boundary) and that can efficiently subsliced.


    The methods of this class will raise IndexError if attempting to access
    a zone outside the current slice or if if the requested slice has its
    first bit located after the last bit.
    """

    # TODO: accept is_binary() in str_
    @typecheck
    def __init__(self, buff: bytes,
                 left: optional(int) = None,
                 right: optional(int) = None,
                 left_bits: optional(int) = None,
                 right_bits: optional(int) = None):
        """Create a new slice
        - buff        the buffer containing the slice
        - left        offset in bytes of the beginning of the slice
        - left_bits    offset in bits  of the beginning of the slice
        - right        offset in bytes of the end of the slice
        - right_bits    offset in bits  of the end of the slice

        A slice is a part of a bytes buffer. It may or may not begin
        and/or end on a byte boudary in the buffer.

        The beginning of the slice is located with the left and
        left_bits arguments and the end is located with the right and
        right_bits arguments.

        These arguments contain an offset value taken from the beginning
        of the buffer if they are positive and from the end of the
        buffer if they are negative.

        Default offsets

          If left & left_bits are ommited, the slice begins at the
          beginning of the buffer. If right & right_bits are ommited,
          the slice begins at the beginning of the buffer.

        Offset combinations

          If both left & left_bits (resp. right & right_bits) have a
          value, then these offset values will be added (note that they
          must have the same sign)

          For example the two following slices will be identical:
              BinarySlice (b"abc", left=1, left_bits=2)
            BinarySlice (b"abc", left_bits=10)
        """
        assert left is None or left_bits is None or (left * left_bits) >= 0  # same sign
        assert right is None or right_bits is None or (right * right_bits) >= 0  # same sign

        if left is None and left_bits is None:
            left_bits = 0

        self.__str = buff
        self.__left = self.__compute_offset(left, left_bits)
        self.__right = self.__compute_offset(right, right_bits)

        if self.__left > self.__right:
            raise IndexError()

    @typecheck
    def same_buffer_as(self, other_slice: this_class) -> bool:
        """Return true if both slices are based on the same buffer

        (usually when one slice is a subslice of the other one)
        """
        return self.__str == other_slice.__str

    @typecheck
    def get_right(self) -> int:
        """Return the right offset (end of the slice) in bits"""
        return self.__right

    @typecheck
    def get_left(self) -> int:
        """Return the left offset (beginning of the slice) in bits"""
        return self.__left

    # length in bytes
    @typecheck
    def __len__(self) -> int:
        """Return the length of the slice in bytes

        If slice does not contain an integer number of bytes (not a
        multiple of 8 bits), the returned result will be the floor
        value. Eg: len(BinarySlice(b"abc", right_bits=10)) == 1
        """
        return (self.__right - self.__left) // 8

    def __bool__(self) -> bool:
        "Return True if the slice is not empty"

        return self.__right != self.__left

    # length in bits
    @typecheck
    def get_bit_length(self) -> int:
        """Return the length of the slice in bits"""
        return self.__right - self.__left

    @typecheck
    def __compute_offset(self, bytes_: optional(int), bits: optional(int), relative: bool = False) -> int:
        """compute a relative offset (typically to generate a subslice)"""
        assert bytes_ is None or bits is None or (bytes_ * bits) >= 0  # same sign

        if relative:
            orig_left, orig_right = self.__left, self.__right
        else:
            orig_left, orig_right = 0, len(self.__str) * 8

        if bytes_ is None:
            if bits is None:
                result = orig_right
            elif bits >= 0:
                result = orig_left + bits
            else:
                result = orig_right + bits
        else:
            if bytes_ >= 0:
                result = orig_left + bytes_ * 8
            elif bytes_ < 0:
                result = orig_right + bytes_ * 8

            if bits:
                result += bits

        if result < orig_left or result > orig_right:
            raise IndexError()

        return result

    @typecheck
    def bit_slice(self, left: int, right: int) -> this_class:
        """Generate a subslice of this slice

        The offsets are given in bits relative to the beginning and end
        of the current slice.
        """
        left = self.__compute_offset(None, left, True)
        right = self.__compute_offset(None, right, True)

        return BinarySlice(self.__str, left_bits=left, right_bits=right)

    @typecheck
    def __getitem__(self, index: either(int, slice)) -> either(int, this_class):
        """Return the value of a byte in the slice (if index is a int)
        or a subslice (if index is a slice)

        The index is given in bytes from the beginning of the slice (if
        positive or null) or from the end of the slice (if negative).

        If a slice is requested, this function will return a BinarySlice
        objet. The offset (index.start and index.stop) are given in
        bytes relative to the beginning and the end of the current
        slice.

        NOTE: index.step is not supported (must be None)
        """

        if type(index) == int:
            return self.get_byte(index * 8)
        else:
            assert index.step is None  # increment not supported

            return BinarySlice(self.__str,
                               left_bits=self.__left if index.start is None else self.__compute_offset(index.start,
                                                                                                       None, True),
                               right_bits=self.__compute_offset(index.stop, None, True)
                               )

    @typecheck
    def get_byte(self, bit_index: int) -> int:
        """Get an individual byte in the string, starting at the offset    bit_index

        bit_index is the offset of the requested byte, given in bits
        from the beginning of the slice (if positive or zero) or from
        the endo of the slice (if negative).
        """
        #        if type (index) == slice:
        #            left  = 0 if index.start == None else self.__compute_offset (index.start, None, True)
        #            right = self.__compute_offset (index.stop , None, True)
        #            step  = index.step*8 if index.step else 8
        #            if left % 8 == 0:
        #                for i in range(0)
        #
        #            print left, right, step
        #            for i in range (*index.indices(self.__right)):
        #            raise

        left = self.__compute_offset(None, bit_index, True)

        # length remaining
        length = self.__right - left
        if not length:
            raise IndexError()

        c = self.__str[left // 8]

        if not left % 8:
            # byte-aligned

            if length < 8:
                # clear the last bits
                c = (c >> (8 - length)) << (8 - length)
        else:
            # NOT byte-aligned
            offset = left % 8

            c = (c & (0xff >> offset)) << offset

            if length > 8 - offset:
                # read a second char
                c |= (self.__str[left // 8 + 1]) >> (8 - offset)

            if length < 8:
                # clear the last bits
                c = (c >> (8 - length)) << (8 - length)
        return c

    @typecheck
    def raw(self) -> bytes:
        """Return the raw bytes string

        This function will return the content of the slice as a raw
        bytes object.

        If the size of the slice is a multiple of 8-bits, then the
        length of the resulting string will be equal to len(self)
        otherwise it will be equal to len(self)+1 and the least
        significant bits of the last byte will be padded with undefined
        values.
        """
        # TODO: optimise this
        return bytes([self.get_byte(i) for i in range(0, self.__right - self.__left, 8)])

    @typecheck
    def as_binary(self) -> is_binary:
        """Return the content of the slice as a binary object"""
        buff = self.raw()
        bit_add = self.get_bit_length() % 8
        return (buff, bit_add) if bit_add else buff

    @typecheck
    def shift_bits(self, bits: int) -> this_class:
        """Create a subslice of the current slice by removing a group of
        bits at the beginning or at the end of the slice.

        This is equivalent to:
        - self.bit_slice (bits, self.get_bit_length()) if bits>=0
        - self.bit_slice (0, bits) if bits<0
        """
        if bits >= 0:
            return BinarySlice(self.__str, left_bits=self.__compute_offset(None, bits, True), right_bits=self.__right)
        else:
            return BinarySlice(self.__str, left_bits=0, right_bits=self.__compute_offset(None, bits, True))

    def __repr__(self):
        """Return a representation of the slice evaluable in python"""
        return "BinarySlice (%s, %d, %d, %d, %d)" % (
            repr(self.__str),
            self.__left // 8,
            self.__right // 8,
            self.__left % 8,
            self.__right % 8
        )


class Data(named.NamedObject):
    """Root class for all Data object

    Data objects are used to represent test data (values and templates).

    see help(core.data) for more details
    """

    __default_name = None

    @typecheck
    def __init__(self, parent: optional(is_data) = None):
        """Constructor

        parent is the parent object of this data. Its type must be
        compatible with self.get_type() (see store_data())
        """

        named.NamedObject.__init__(self, Data.__default_name)
        self.__frozen = False
        self.__parent = None

        if parent:
            self.set_parent(parent)

    @staticmethod
    @contextmanager
    def disable_name_resolution():
        """Return a context that disables the automatic name resolution
        done by NamedObject.__init__

        All Data object have a __name__ attribute. If after its
        creation, the object is immediately store into a variable, then
        the __name__ attribute is set to the name of this variable
        (otherwise it is set to "(anon)".

        This process is convenient for tracking the name data objects,
        however it is a little time consuming and can induce significant
        delays when lots of objects are created (especially when
        encoding/decoding messages).

        This function returns a context that disables the name
        resolution. It is automatically during the execution of a
        testcase.

            >>> IntValue(12).__name__
            '(anon)'

            >>> three = IntValue(3)
            >>> three.__name__
            'three'

            >>> three_bis = three
            >>> three_bis.__name__
            'three'

            >>> with Data.disable_name_resolution():
            ...    four = IntValue(4)
            ...    print (four.__name__)
            ...
            (anon)
        """
        previous_value = Data.__default_name
        Data.__default_name = "(anon)"
        try:
            yield
        finally:
            Data.__default_name = previous_value

    @typecheck
    def set_parent(self, parent: is_data):
        """Set the parent of this data object

        Its type must be compatible with self.get_type() (see store_data())

        This function may be used only once on one object.

        NOTE: this is the only function that can alter a Data object. It
        may be removed in the future
        """
        parent = store_data(parent,
                            self.get_type())  # FIXME: the conversion might be too strict (especially with a subclass)  -> should we allow the child to be a subclass of the parent ?

        assert not self.is_frozen()
        if self.__parent is not None:
            raise exceptions.Error("This data has already one parent data")

        parent.freeze()
        self.__parent = parent

    def get_parent(self) -> optional(is_data):
        """Get the parent object this object derives from"""
        return self.__parent

    def get_type(self) -> type:
        """Return the type of Data contained in this object.

        This function returns the type of the object with the same
        meaning as the global get_type() function. The result is a
        subclass of Value.

        NOTE: the result is not necessarily equals to type(self) (this
        is true only for Value instances, Template instances store their
        type in an attribute)
        """
        raise exceptions.NotImplemented()

    @typecheck
    def datas(self) -> iterable:
        """Return an iterator on this object and its parents"""
        d = self
        while d:
            yield d
            d = d.__parent

    def freeze(self):
        """Freeze this object (make it immutable).

        NOTE: may be removed in the future
        """
        self.__frozen = True

    def is_frozen(self):
        """Return true if this object is immutable.

        NOTE: may be removed in the future
        """
        return self.__frozen

    def describe(self, desc):
        """Fill a MessageDescription object with the description of this data.

        The default implementation returns the type of the data.

        It may be reimplemented in order to give more detailed
        information (eg. source/destination addresses)

        The function return True if a description could be successfully
        extracted.
        """
        assert isinstance(desc, MessageDescription)

        desc.info = self.get_type().__name__
        return True

    @typecheck
    def match(self, value: optional(is_flat_value), mismatch_list: optional(list) = None) -> bool:
        """Return True if the argument 'value' matches the current data.

        'value' will match 'self' if all the following conditions are
        fulfilled:
         - its type is compatible ('value' is an instance of self.get_type())
         - in case 'self' derives from a parent data, 'value' must match
           this data
         - it fulfils type-specific checks implemented in self._match()

        The mismatch_list can be used to get detailed information in
        case of mismatch. If used, it must be initialised to an empty
        list which will be populated with a set of Mismatch objects.

        NOTE: This function should not be reimplemented. Specialised
        match behaviour should be implemented or reimplemented in the
        _match() member function instead.
        """

        # TODO: question: is it possible for a template to derive from multiple types ?
        #    (eg. combination of templates) -> how to handle that ???

        result = True

        if mismatch_list is not None:
            # just used to ensure that the mismatch_list is populated if and only if we have a mismatch
            initial_mismatch_list_length = len(mismatch_list)

        if value is None:
            result = False
            if mismatch_list is not None:
                mismatch_list.append(TypeMismatch(value, self))
        else:
            # FIXME: should we have automatic conversion to type(self) ?
            value = as_data(value)

            if self.__parent is None:
                # top template
                # -> first check that the data is compatible with this type

                # FIXME: should't we do this in the constructor directly ?
                if not self.get_type().is_valid_value(value):
                    result = False
                    if mismatch_list is not None:
                        mismatch_list.append(TypeMismatch(value, self))

            else:
                # child template
                # -> first check with the parent template
                if not self.__parent.match(value, mismatch_list):
                    if mismatch_list is not None:
                        # will not change the result,
                        # but this will give a more accurate mismatchlist
                        if self.get_type().is_valid_value(value):
                            self._match(value, mismatch_list)
                    result = False

            # run type-specific checks
            if result:
                result = self._match(value, mismatch_list)

        if mismatch_list is not None:
            # when _match() returns False, there must be at least one difference
            have_no_new_mismatches_entries = len(mismatch_list) == initial_mismatch_list_length

            assert have_no_new_mismatches_entries == result  # at least one mismatch entry must be created in case of mismatch

        return result

    @typecheck
    def _match(self, value: is_flat_value, mismatch_list: optional(list) = None) -> bool:
        """Specialised match method to be reimplemented in inherited
        classes

        The match() methods performs generic checks that are common to
        every Data classes:
            - check that value is an instance of self.get_type()
            - check that the value matches the parents of self (if
              any)

        This function is expected to check if 'value' matches the
        pattern described in 'self' and return True if the match is
        success.

        In case of mismatch, the function must return False and append a
        Mismatch object into the mismatch list (if present) in order to
        give details about the mismatch.
        """
        raise exceptions.NotImplemented()

    @typecheck
    def display(self, indent=0, output: with_attr("write") = sys.stdout):
        """Display the current data in a multi-line human readable way.

        Parameters:
         - indent: number of blank spaces to prepend in the result
                (for representing value encapsulation)
         - output: output stream (default: sys.stdout)
        """
        return self._display(indent, output)

    # TODO: we should keep a reference to the original template
    # FIXME: exceptions thrown by this function may be difficult to handle in practice
    def flatten(*datas) -> is_flat_value:
        """Generate a flat value from this data object and other data

        The list of data objects is split in two groups:
         - Value objects
         - Template objects

        If there is no value object, None is returned.


        In order to be flattened, the values must all have exactly the
        same type (otherwise an exception is raised).

        Then the specialised _flatten() method is called to generate the
        actual flattened value.

        The resulting value is compared with the template objects listed
        above. If there is any mismatch, then an exception is returned.


        NOTE: this function should not be reimplemented. You should
        reimplement _flatten() instead
        """

        # First we separate values and templates
        values = []
        templates = []
        for d in datas:
            d.freeze()

            if isinstance(d, Template):
                templates.append(d)
            else:
                assert isinstance(d, Value)
                values.append(d)

        # we must have at least one value
        if len(values) == 0:
            # FIXME: maybe raise an error if there are templates defined but no values
            return None

        # check that all are of compatible types
        value_class = type(values[0])

        for d in datas:
            if d.get_type() != value_class:  # FIXME: might be too strict
                raise exceptions.Error("incompatible types cannot be flattened together " + str(datas))

        # flatten the values
        result = value_class._flatten(values)

        # check that the result complies with the templates
        for t in templates:
            if not t.match(result):
                raise exceptions.Error("flattened value does not comply with one of its templates")

        return result

    def is_flat(self):
        """Return True if this data is flat

        A flat data is a Value object that has no parents and in case of
        structured objects, a value that contains only flat values.

        This function checks that this data has no parent and calls
        self._is_flat() for specialised checks.


        NOTE: this function should not be reimplemented. You should
        reimplement _is_flat() instead
        """
        return (self.__parent is None) and self._is_flat()

    @typecheck
    def __contains__(self, value: is_value) -> bool:
        '''shortcut for the match() method'''

        return self.match(value)

    @named.skip_parent_var_name
    @typecheck
    def __call__(self, *k, **kw) -> this_class:
        """Generate a derived value from this data object.

        The parameters *k, **kw are forwarded to the constructor of the
        Value class. The parent of the resulting object will be the
        current object.
        """
        v = type(self)(*k, **kw)
        v.set_parent(self)
        return v

    def __repr__(self):
        """Return a python-evaluable source representation of the data

        This function handles the inheritance aspects (setting the
        parent) and calls _repr() to generate the type-specific
        instantiation code.

        NOTE: this function should not be reimplemented. You should
        reimplement _repr() instead
        """

        if hasattr(self, "_repr"):

            if self.get_parent() is None:
                return self._repr()
            else:
                return "%s.set_parent(%s)" % (self._repr(), repr(self.get_parent()))
        else:
            # TODO: issue a warning
            return object.__repr__(self)


# TODO: maybe remove this function and use store_data instead
@typecheck
def as_data(arg: is_data, type_=None):
    """Get the arguement as a Data object

    The argument must either be an instance of Data or a python primitive value.

    If it is a primitive value, the function will generate a PrimitiveValue
    instance encapsulationg this primitive value.


    If a type is given, the function will try to convert the result into
    this type (only allowed for primitive types for the moment).
    """
    assert (not type_) or issubclass(type_, Value)

    if type_ in (None, Value):
        # no hint type, just convert the value if needed

        if isinstance(arg, Data):
            return arg
        else:
            # assume that it is a primitive type
            return primitive.PrimitiveValue.new(arg)
    else:
        # the type of the result must be the same as type_

        if isinstance(arg, type_):
            # nothing to do
            return arg
        else:
            # TODO: remove this case (it is now handled in store_data())
            try:
                # our data has a different type
                #
                # let's try to convert it

                assert issubclass(type_, primitive.PrimitiveValue)  # TODO: generalise it to all types ?
                return type_(arg)

            except Exception as e:
                raise exceptions.BadConversion(arg, type_, e)


# Return a data object
# FIXME:we check that the data is of the right type. Shall we also check that the value is allowed (eg. Subtype -> better to be done in the constructor)
# FIXME: we could remove sepecial treatment form Omit() here and use a kind of Union instead
def store_data(arg, type_=None, none_is_allowed=False, omit_is_allowed=False):
    """Prepare a data to be stored.

    store_data() is expected to be used in any place it is necessary to
    store a data object (especially in the implementation of structured
    values and templates). Actually this function is mostly implicitely used
    when creating structures Value objects. For example if one writes:
        IPv6 (hl=1, src="fe80::4242")
    the constructor of the IPv6 class will implicitely call:
        store_data (1, UInt8)    for the hl field
        store_data ("fe80::4242", IPv6Address)    for the src field
    in order to convert the input into the desired types (the type of the hl
    field is UInt8 and the type of src is IPv6Address).


    The caller may specify a target type. If the argument is an instance of
    this type, then it is used as is. Otherwise the function will try to
    build an instance of this type and will forward the argument to its
    constructor (in that case 'arg' can be anything that is accepted by the
    constructor). This is used to generate a value from a different input
    format, for example an IPv6 address is a subclass of BytesValue but is
    constructor is specialised and accepts the input value to be formatted
    as a human readable string. Thus is is possible to write:
        store_data ("ff02::1", IPv6Address)
    instead of:
        store_data (b"\xff\2\0\0\0\0\0\0\0\0\0\0\0\0\0\1", IPv6Address)

    The caller may also specify whether it accepts storing the two special
    objects:
    - None        -> undefined value
    - Omit()    -> omitted value (to represent optional fields)
    By default these two values are not allowed and the function will raise
    an exception.

    store_data() performs several important things:

    - In the case arg is equals to None it ensures that none_is_allowed is
      set and return None (otherwise it throws an exception)

    - If arg is a Template, then is ensures that it's type is compatible
      with the requested type (if given)

    - Otherwise it is assumed that the result will be a Value object
      * if no target type is given, then the function will used the input
        argument as is if it is already an instance of Value, otherwise
        it will try to convert it, assuming that it is a primitive
        python type.
      * if a target type is given, then the function uses the argument as is
        if it is an instance of this type, otherwise it will try to build
        an instance of this target type and forward the argument to its
        constructor.

    - Before returning, the Data result is frozen by calling its .freeze()
      method (Data objects are expected to be immutables, if they contain
      other objects, those must also be immutables) FIXME: this may no
      longer be needed in the future
    """
    assert type_ is None or issubclass(type_, Value)

    if arg is None:
        if none_is_allowed:
            # FIXME: it could be better to return some kind of NoneValue object here
            return None
        else:
            raise exceptions.Error("Expected a real value ('None' is not allowed)")

    if isinstance(arg, Template):
        # TODO: clarify this requirement (would never match otherwise)
        # TODO: enforce value vs. type checking in (PacketValue, ListValue, ...)._match()
        # FIXME: how to handle automatic conversions w/ issubclass (arg.get_type(), type_)
        assert type_ is None or issubclass(type_, arg.get_type()) or issubclass(arg.get_type(), type_) or (
            issubclass(arg.get_type(), Omit) and omit_is_allowed)

        data = arg

    else:
        # consider it is going to be a value
        if type_ in (None, Value):
            # no type -> just create a Value object to represent it
            data = as_data(arg)
        else:
            if omit_is_allowed and isinstance(arg, Omit):
                data = arg
            else:
                # convert to the desired type if needed
                try:
                    data = arg if isinstance(arg, type_) else type_(arg)
                except Exception as e:
                    raise exceptions.BadConversion(arg, type_, e)
    # make it immutable
    data.freeze()

    return data


@named.skip_parent_var_name
def pack(*msg_list):
    """This is a shortcut function to encapsulate datas into other datas. It
    makes a recursive call to the pack() methods of each container.

    For example:
        pack (Ethernet(), IPv6(), UDP(), "blahblah")
    is roughly equivalent to:
        Ethernet (pl=IPv6 (pl=UDP (pl="blahblah")))
    and to:
        Ethernet.pack (IPv6.pack (UDP.pack ("blahblah")))
    """
    assert len(msg_list)

    next = None
    # make an object that derives from each of the objects in the list
    # TODO: find a way to avoid duplicating anonymous objects ???
    for msg in reversed(msg_list):
        assert is_data(msg)

        if next is not None:
            next = as_data(msg).pack(next)
        else:
            next = as_data(msg)

    # set a new name for the resulting template
    candidate = named.get_parent_var_name()
    if candidate != "(anon)":  # FIXME: shouldn't this be systematic ? (we should not copy the name of the parent data)
        next.__name__ = candidate

    return next


class Value(Data):
    """Abstract base class for all object representing values.

    see help(Data)
    """

    def get_type(self):
        """Return the type (subclass of Value) associated to this data object

        For a value object, this function returns type(self)
        """
        return type(self)

    @classmethod
    def is_valid_value(cls, value: is_value):
        """Return true if the given value is compatible with the current type"""

        # FIXME: this may be too strict
        #    eg:    class U8 (metaclass = SubtypeClass (Range (int, 0, 255))):
        #            pass
        #        IntValue.is_valid_value (U8 (4))    ->  True
        #        U8.is_valid_value  (U8 (4))        ->  True
        #        U8.is_valid_value  (IntValue (4))    -> *False*
        #        U8.is_valid_value  (4)            -> *False*
        #
        #    should we implement automatic conversions ?
        #    (the same issue is also present in Data.match())
        #
        return isinstance(as_data(value), cls)

    def build_message(self):
        """Build a message from this value.

        This function is recursively called in case of structured
        values. It does not return a Message object but a tuple
        (is_flat_value, is_binary) that contains the built value and its
        binary representation of the Message object to be initialised.

        The value returned in the result may differ from self because
        the _build_message() may fill undefined fields with some default
        or some automatically computed values (eg. a checksum).

        This function only implement some sanity checks. The actual
        implementation is provided by the method _build_message().
        """
        assert self.is_flat()

        result = self._build_message()

        assert is_flatvalue_binary(result)
        assert result[0].get_type() == self.get_type()

        result[0].freeze()

        return result

    # TODO: clarify this function
    def find_type(self, t):
        raise KeyError

    def _build_message(self: is_flat_value) -> is_flatvalue_binary:
        """Implementation of the message encoder.

        This function is expected to be reimplemented.
        """
        raise exceptions.NotImplemented()

    @classmethod
    @typecheck
    def decode_message(cls, bin_slice: BinarySlice) -> is_flatvalue_binslice:
        """Decode a message of the current type from the content of a
        binary slice.

        This function is recursively called in case of structured
        values. It does not return a Message object but a tuple
        (is_flat_value, binary_slice) that contains the decoded value
        and the new slice to be used to next values. Decoding is usually
        done from left to right (starting from the beginning of the
        slice) but there is no obligation to follow this order in the
        implementation.

        At the end of the whole message decoding, the resulting slice is
        expected to be empty (otherwise Message's constructor will raise
        an exception).

        This function only implement some sanity checks. The actual
        implementation is provided by the method _decode_message().
        """

        #        assert issubclass (cls.__value_class, Value) # TODO: report this error

        value, bin_slice = cls._decode_message(bin_slice)

        # TODO: catch exceptions so as to help locating the errors
        # TODO: freeze the message after decoding ?

        # convert the result if needed
        result = as_data(value, cls), bin_slice

        # some sanity checks
        assert is_flatvalue_binslice(result)
        assert isinstance(result[0], cls)

        return result

    @classmethod
    def _decode_message(cls, bin_slice):
        """Default implementation of the message decoder.

        In the case cls is equals to the Value class, this function will
        by default decode the message as a BytesValue. Otherwise an
        exception is raised.

        This function is expected to be reimplemented.
        """

        if cls != Value:
            raise exceptions.NotImplemented()

        # by default if the field type is just 'Value' (typically for a Payload)
        # we decode as bytes
        return primitive.BytesValue._decode_message(bin_slice)


class Omit(Value):
    """Special type to be used in structured types where fields can
    optionally contain no value

    Assigning 'Omit()' to a field is different from assigning 'None'. The
    former is truly considered as a Value object whereas the latter means
    that there is no value defined.

    Thus:
     - Value.build_message() will not replace it with a default value
     - Value.match() will expect that the compared value is also equals to
       Omit() (whereas if the pattern is None, any value will be accepted)

    Omit() is encoded and decoded as an empty string.
    """

    def __new__(cls):
        """Singleton constructor"""

        # singleton
        cls.__instance = super().__new__(Omit)

        def new(cls):
            return cls.__instance

        cls.__new__ = new
        return cls.__instance

    def _build_message(self):
        return self, b""

    @classmethod
    def _decode_message(cls, bin_slice):
        return cls(), bin_slice

    def _is_flat(self):
        return True

    @classmethod
    def _flatten(cls, values):
        if all(v is cls.__instance for v in values):
            return cls.__instance
        else:
            raise exceptions.Error("Cannot flatten Omit() with other values")

    def _match(self, value, diff):
        assert self is value
        return True

    def __str__(self):
        return "(omit)"

    def __repr__(self):
        return "Omit()"


class Template(Data):
    """Abstract base class for representing templates

    see help(Data)
    """

    @typecheck
    def __init__(self, type_or_parent: either(is_type, is_data)):
        """Constructor

        type_or_parent contains either:
         - the type of this template (if this template has no parent
           data)
         - the parent data for this template (in that case, the type of
           this new template is inferred from the type of the parent)
        """

        self.__type = get_type(type_or_parent)

        if is_type(type_or_parent):
            Data.__init__(self)
        else:
            Data.__init__(self, type_or_parent)

    def get_type(self):
        """Return the type of this template

        The result is a subclass of Value (corresponding to the type
        used for initialising this template)
        """
        return self.__type

    def _is_flat(self):
        return False

    def _match(self, value, mismatch_list):
        """Specialised implementation of _match for templates

        This method is an intermediate method for implementing
        specialised matches for templates.

        In case of templates mismatch, we return only one mismatch
        object TemplateMismatch for the whole template. This function
        generates this object if needed upon the call to the method
        _template_match() if needed.
        """
        result = self._template_match(value)
        if not result and mismatch_list is not None:
            mismatch_list.append(TemplateMismatch(value, self))
        return result

    def _template_match(self, value: is_flat_value) -> bool:
        """Specialised implementation of _match to be reimplement for
        each template class

        This method shall return true if the value matches the template.
        """

        raise exceptions.NotImplemented()

    def store_data(self, data, none_is_allowed=False):
        """Shortcut to call store_data using this template's type.

        This is equivalent to calling global function:
            store_data (data, self.get_type(), none_is_allowed=...)

        By default None is not allowed.

        Omit() is never allowed (FIXME: maybe we should authorise it?)
        """
        return store_data(data, self.__type, none_is_allowed)


# NOTE: some data may appear multiple times in the tree
#    this does normally not apply to values since they are supposed to be flat
#    (but this may not be sufficent)
class Mismatch:
    """Base mismatch object to document match failures.

    If a list is given as parameter (mismatch_list) to the function
    Data.match(), it will be populated with one or more Mismatch object in
    case of failure.

    The order of the mismatches in the list does matter. In case of
    structured values, mismatches detected in its fields should be placed
    after the mismatch detected on the structured value itself (eg. length
    mismatch). This is important to ease the parsing of the mismatch list
    later.
    """

    @typecheck
    def __init__(self, value: optional(is_value), pattern: is_data):
        """Initialise the Mismatch object

         - 'pattern' it the pattern that was expected
         - 'values' is the value that failed to match the pattern
        """
        # FIXME: these attributes should not be modifiable
        self.value = value
        self.pattern = pattern

    def describe_value(self, describe_func=None) -> str:
        """Return a textual description of the input value

        It may be reimplemented

        Describe_func is an optional callable object that takes
        a value of self's type and return a textual description
        for this value.
        """
        txt = describe_func(self.value) if describe_func else None

        if txt is None:
            return str(self.value)
        else:
            return "%s (%s)" % (self.value, txt)

    def describe_expected(self, describe_func=None) -> str:
        """Return a textual description of the expected pattern

        It may be reimplemented

        Describe_func is an optional callable object that takes
        a value of self's type and return a textual description
        for this value.
        """
        txt = describe_func(self.pattern) if (describe_func and is_value(self.pattern)) else None

        if txt is None:
            return str(self.pattern)
        else:
            return "%s (%s)" % (self.pattern, txt)

    def describe_full(self, describe_func=None) -> str:
        """Return a string describing the mismatch

        The default implementation returns the name of the type of self
        and the result of describe_expected() and describe_value()

        Describe_func is an optional callable object that takes
        a value of self's type and return a textual description
        for this value.
        """
        return "%s: expected %s, got %s" % (type(self).__name__, self.describe_expected(), self.describe_value())

    def __str__(self):  # TODO: rename it as describe_brief and make describe_full the default
        """Return a string describing the mismatch

        The default implementation returns the name of the type of self
        and the result of describe_expected()
        """
        return "%s: expected %s" % (type(self).__name__, self.describe_expected())


class ValueMismatch(Mismatch):
    """Value mismatch

    The pattern is a different value.
    """
    pass


class TypeMismatch(Mismatch):
    """Type mismatch

    The pattern and the value have different types.
    """

    def describe_value(self, describe_func=None) -> str:
        return type(self.value).__name__

    def describe_expected(self, describe_func=None) -> str:
        return self.pattern.get_type().__name__


class VariantMismatch(Mismatch):
    """Variant mismatch

    The pattern and the value are derived from the same type, but have a
    different variant.
    """

    def __init__(self, v, p):
        Mismatch.__init__(self, v, p)
        assert isinstance(p, packet.PacketValue)

    def describe_value(self, describe_func=None) -> str:
        return type(self.value).__name__

    def describe_expected(self, describe_func=None) -> str:
        return self.pattern.get_variant().__name__


class LengthMismatch(Mismatch):
    """Length mismatch

    The value and the pattern have a different length
    """

    def describe_value(self, describe_func=None) -> str:
        return "%d elements" % len(self.value)

    def describe_expected(self, describe_func=None) -> str:
        return "%d elements" % len(self.pattern)


class TemplateMismatch(Mismatch):
    """Template mismatch

    The value did not match the template
    """
    pass


class DifferenceList(list):
    # (type, callable (value) -> yields (name, value))
    # TODO: design a standard interface in Data to introspect structured datas
    __handlers = {}

    def __init__(self, value):
        self.__value = value

    def callback(self, path, mismatch):
        print("%s: %s" % (".".join(path), mismatch.describe_full()))

    @classmethod
    @typecheck
    def set_handler(cls, typ: is_type, handler: callable):
        cls.__handlers[get_type(typ)] = handler

    @typecheck
    def describe(self, callback: callable = callback):

        path = []
        walkers = []
        value = self.__value

        walker = None
        path_elem = type(value).__name__
        for mismatch in self:

            while True:
                if mismatch.value is value:  # how about the case value==None ?
                    path.append(path_elem)
                    callback(path, mismatch, describe)
                    path.pop()

                    break  # next mismatch

                for typ, handler in self.__handlers.items():
                    if isinstance(value, typ):
                        walkers.append(walker)
                        walker = handler(value)
                        if path_elem is not None:
                            path.append(path_elem)

                            assert isinstance(path[-1], str)
                        break
                while True:
                    if walker is None:
                        callback(["(unknown)"], mismatch, None)
                        break

                    try:
                        path_elem, value, describe = next(walker)
                        break
                    except StopIteration:
                        path.pop()
                        walker = walkers.pop()

                if walker is None:
                    break


class Bidict:
    """An implementation of bi-directionnal dictionnary

    Bidict is a dictionnary that can be browsed by key or by value.

    The syntax for accessing its content by value is to use __getitem__ or
    __setitem__ with a slice whose stop value it the requested value.

    By default duplicate entries are not allowed. Thus if a key (or value)
    is overwritten, the previous key+value entry is deleted.

    Exemple:
        >>> bd=Bidict()
        >>> bd[1]=2
        >>> bd[1]
        2
        >>> bd[:2]
        1
        >>> bd[:2]=3
        >>> bd[3]
        2
        >>> bd[1]
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
          File "/home/abaire/git/ttproto/core/data.py", line 1657, in __getitem__
            return self._fwd[item]
        KeyError: 1
        >>>
    """

    @typecheck
    def __init__(self, init_entries: optional(either(dict, this_class)) = None, allow_duplicates=False):
        """Initialise the dictionnary

        Parameters:
         - init_entries: a dict or Bidict to populate this new Bidict
         - allow_duplicates: if True, duplicate entries will be allowed
           in the Bidict (default is: not allowed)
        """
        self._fwd = {}
        self._bwd = {}
        self.__allow_duplicates = allow_duplicates

        if init_entries:
            self.update(init_entries)

    def __getitem__(self, item):
        """Acces one element in the dictionnary

        Forward resolution (key -> value):

            bidict[key] returns the associated value


        Backward resolution (value -> key):

            bidict[:value] returns the associated Key


        The function raises KeyError if the value or key is not found.
        """

        if isinstance(item, slice):
            assert item.step is None

            if item.start is not None:
                # forward resolution
                assert item.stop is None

                return self._fwd[item.start]
            else:
                # backward resolution
                assert item.stop is not None

                return self._bwd[item.stop]
        else:
            # forward resolution
            return self._fwd[item]

    def __setitem__(self, item, value):
        """Add one entry in the Bidict

        Like __getitem__, this function can be used in the two
        directions:
            bidict[key] = value        (forward)
        or
            bidict[:value] = key        (backward)


        If the Bidict does not allow duplicates (see __init__), the two
        variants above will have the same result (and the previous
        entries with the same value or key will be deleted)


        When duplicates are allowed, the previous entries are not
        deleted.
        """
        if isinstance(item, slice):
            assert item.step is None

            if item.start is not None:
                # forward assignment
                assert item.stop is None

                self._set(item.start, value)
            else:
                # backward assignment
                assert item.stop is not None

                self._set(value, item.stop)
        else:
            # forward assignment
            self._set(item, value)

    def _set(self, key, value):
        if not self.__allow_duplicates:
            # remove previous associations
            if key in self._fwd:
                del self._bwd[self._fwd[key]]
            if value in self._bwd:
                del self._fwd[self._bwd[value]]

        self._fwd[key] = value
        self._bwd[value] = key

    def __iter__(self):
        """Return an iterator on the keys of the bidict"""
        return iter(self._fwd)

    def items(self):
        """Return an item iterator (key, value) using the forward dictionnary"""

        return self._fwd.items()

    def update(self, src: either(dict, this_class)):
        """Update the bidict with a set of entries"""

        for key, value in src.items():
            self._set(key, value)

    def __bool__(self):
        """Return True if the bidict is not empty"""
        return bool(self._fwd) or bool(self._bwd)

    def __repr__(self):
        return "%s({%s})" % (
            type(self).__name__,
            ", ".join("%s: %s" % (repr(k), repr(v)) for k, v in self.items())
        )


class BidictValueType(Bidict):
    """A bi-dictionnary specifically for doing value type resolution

    This bidict class is used for associating a value to a type, this is
    needed encoding and decoding messages. Typically for the TLV-like
    formats (type-length-value).

    When building a message, the "type" field is filled according to the
    type of the "value" field (eg. an IPv6 header containing an ICMPv6
    message will have its NextHeader field to be set to 58).

    When decoding a message, the content of the "type" field is used to know
    what is the expected type of the value (and thus to know which decoder
    shall be called) (eg. an IPv6 header with a NextHeder set to 58 is
    expected to contain an ICMPv6 message).


    This specialised bidict has some differences compared to Bidict:
    - the keys are Value objects   (canonicalised with as_data())
    - the values are types (subclasses of Value)  (canonicalised with
      get_type())
    - the bidict has a default value and a default type, thus if a lookup
      fails, the bidict will not raise any exception but will return the
      default value or type
    - when looking up for a type, the implementation will also look up its
      base classes if the type is not found in the bidict, it will return
      the first result it finds in type.__mro__.
    """

    @typecheck
    def __init__(self, default_value: optional(is_value), default_type: optional(is_type), *k, **kw):
        super().__init__(*k, **kw)
        self.__default_value = None if default_value is None else as_data(default_value)
        self.__default_type = None if default_type is None else get_type(default_type)

    @typecheck
    def _set(self, value: is_value, type_: is_type):
        super()._set(as_data(value), get_type(type_))

    def __getitem__(self, item):
        if not isinstance(item, slice) or item.stop is None:
            # forward resolution (Value -> Type)
            if isinstance(item, slice):
                assert item.stop is None
                assert item.step is None

                item = item.start
            try:
                return super().__getitem__(as_data(item))
            except KeyError:
                if self.__default_type is not None:
                    return self.__default_type
                else:
                    raise
        else:
            # backward resolution (Type -> Value)
            # try with all types in __mro__
            assert isinstance(item.stop, type)
            for t in get_type(item.stop).__mro__:
                try:
                    return super().__getitem__(slice(None, t, None))
                except KeyError:
                    pass
            if self.__default_value is not None:
                return self.__default_value
            else:
                raise KeyError(item.stop)

    def __repr__(self):
        return "%s(%s, %s, {%s})" % (
            type(self).__name__,
            repr(self.__default_value),
            repr(self.__default_type),
            ", ".join("%s: %s" % (repr(k), repr(v)) for k, v in self.items())
        )


class Message:
    """A class to represent a message exchanged between two entities.

    A message has two representations:
     - a value representation (which is an instance of the Value class)
     - a binary representation (which is a binary object (see is_binary())

    see help(core.data)
    """

    @typecheck
    def __init__(self, data_or_binary: either(is_data, is_binary), expected_type=None):
        """Initialise the message with either the value or the binary
        representation.

        If the input is a value, then the message is built from the
        result of Value.built_message()

        If the input is a binary object, then the expected type must be
        provided too. Then the binary is decoded using
        expected_type.decode_message() and the message is built from the
        result.    In case of decoding error, a DecodeError exception is
        raised.
        """
        if expected_type is None:
            # Build from a value
            assert is_data(data_or_binary)

            self.__value, self.__bin = as_data(data_or_binary).flatten().build_message()

        else:
            try:
                # Build from a binary message
                assert is_binary(data_or_binary)

                self.__bin = data_or_binary
                self.__value, binslice = expected_type.decode_message(BinarySlice(data_or_binary))

                self.__value.freeze()  # TODO: move this into decode_message ?

                if binslice.get_bit_length():
                    # NOTE: Put this as comment otherwise the process is too long
                    # self.__value.display()
                    raise exceptions.Error("Buffer not fully decoded (%d bits remaining)" % binslice.get_bit_length())
            except Exception as e:
                raise exceptions.DecodeError(data_or_binary, expected_type, e)

        self.__description = None

        assert is_binary(self.__bin)
        assert isinstance(self.__value, Value)
        assert self.__value.is_flat()
        assert self.__value.is_frozen()

    def get_value(self) -> Value:
        """Return the value representation of the message"""
        return self.__value

    @typecheck
    def get_binary(self) -> is_binary:
        """Return the binary representation of the message"""
        return self.__bin

    def get_description(self):
        """Return a MessageDescription object with summary information
        about the message"""

        if not self.__description:
            self.__description = MessageDescription(self)
        return self.__description

    def __str__(self):
        return "Message{%s, %s}" % (self.__value, self.__bin)

    def summary(self):
        """Return summary information about the message

        This is equivalent to str(self.get_description())
        """
        return str(self.get_description())

    def display(self, indent=0, output: with_attr("write") = sys.stdout):
        """Output a multi-line textual description of the message (value
        & binary)"""
        self.__value.display(indent, output)
        pfx = " " * indent
        print("%sEncoded as:" % pfx, file=output)
        b = self.__bin
        if isinstance(b, bytes):
            remainder = None
        else:
            remainder = b[1]
            b = b[0]
        for i in range(0, len(b), 16):
            print("%s    %s  %s" % (
                pfx,
                " ".join(format(c, "02x") for c in b[i:i + 8]),
                " ".join(format(c, "02x") for c in b[i + 8:i + 16])), file=output)
        if remainder:
            print("%s(last byte truncated at bit %d)" % (pfx, remainder), file=output)


class MessageDescription:
    """A class for describing a message

    This class is used for extracting specific informations about a message.

    The supported attributes are:
     - src        source address
     - dst        destination address
     - hw_src    hardware source address
     - hw_dst    hardware destination address
     - info        short summary about the content of the message

    At initialisation time, the MessageDescription class calls the
    describe() method on the value representation of the message.

    This method can be reimplemented in each type of message in order to
    fill specific informations.

    Attributes can be accessed using the dot notation:
        >>> md = MessageDescription(Message(IPv6(src="fe80::1234", dst="fe80::5678")/UDP(sport=1024, dport=53)/"test"))
        >>> md.src
        'fe80::1234'
        >>> md.hw_src
        ''
        >>> str(md)
        '[fe80::1234                       -> fe80::5678                      ] UDP 1024 -> domain'
    """

    # Added "src_port" and "dst_port" after e6dc5f5b0b77c6bb987374e757e14e9362191c1e
    __attrs = ("hw_src", "src", "hw_dst", "dst", "src_port", "dst_port", "info")

    @typecheck
    def __init__(self, message: Message):
        """Initialise the description from the given message"""
        message.get_value().describe(self)

    def __setattr__(self, name, value):
        """Update the value of one attribute

        NOTE: the attribute is updated if the given value is not None
        """
        assert name in self.__attrs
        if value is not None:
            super(MessageDescription, self).__setattr__(name, as_data(value))

    def __getattr__(self, name):
        """Get the content of an attribute

        If the attribute has no value set, then the function returns an
        empty string.
        """
        assert name in self.__attrs
        return ""

    def __str__(self):
        """Return a one-line summary of the message

        The format is the following:
            [src -> dst] info
        """
        src, dst = self.src, self.dst
        if (not src) and (not dst):
            src, dst = self.hw_src, self.hw_dst

        return "[%-2s -> %-2s] %s" % (src, dst, str(self.info))


from ttproto.core import primitive, packet



from ..common import *


class TD_COAP_CORE_02(CoAPTestCase):
    """
---
TD_COAP_CORE_02:
    cfg: CoAP_CFG_BASIC
    obj: Perform DELETE transaction (CON mode)
    pre: Server offers a /test resource that handles DELETE
    ref: '[COAP] 5.8.4, 1.2, 2.1, 2.2, 3.1'
    seq:
    -   s:
        - 'Client is requested to send a DELETE request with:'
        -   - Type = 0 (CON)
            - Code = 4 (DELETE)
    -   c:
        - 'The request sent by the client contains:'
        -   - Type=0 and Code=4
            - "Client-generated Message ID (\u2794 CMID)"
            - "Client-generated Token (\u2794 CTOK)"
            - Uri-Path option "test"
    -   c:
        - 'Server sends response containing:'
        -   - Code = 2.02 (Deleted)
            - Message ID = CMID, Token = CTOK
            - Content-format option if payload non-empty
            - Empty or non-empty Payload
    -   v: Client displays the received information
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [CoAP(type='con', code='delete', opt=Opt(CoAPOptionUriPath("test")))]

    def run(self):

        verdict_if_none = "inconclusive"

        while self.match('client', CoAP(type='con', code='delete', opt=self.uri('/test')), verdict_if_none):
            CMID = self.coap['mid']
            CTOK = self.coap['tok']

            self.next()

            if self.match('server', CoAP(code=2.02, mid=CMID, tok=CTOK)):
                pass  # Pass cases, nothing to do.
            elif self.match('server', CoAP(mid=CMID, tok=CTOK), None):
                code = self.coap['code']
                # Fail cases : when code is 2.01/2.03/2.04/2.05
                if code == 65 or code == 67 or code == 68 or code == 69:
                    self.set_verdict("fail", "Server responded to a deletion request with " + str(code))
                else:  # Inconclusive cases.
                    # Other code 4.xx/5.xx, e.g various errors.
                    # We cannot trow a Fail verdict here because even if it do
                    # not correspond to a check step, it does not violate the RFC
                    # either.
                    self.set_verdict("inconclusive", "Server responded to a deletion request with " + str(code))

            if self.match('server', CoAP(pl=Not(b'')), None):
                self.match('server', CoAP(opt=Opt(CoAPOptionContentFormat())), 'fail')

            verdict_if_none = None
            self.next(True)



#!/usr/bin/env python3

#from typecheck3000 import typecheck, optional, either, iterable, tuple_of, list_of, dict_of, with_attr

"""


assert IPv6Address ("2001:DB8::1") == b' \x01\r\xb8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01'
assert IPv6Address ("fe80::212:34ff:fe56:7890") == b'\xfe\x80\x00\x00\x00\x00\x00\x00\x02\x124\xff\xfeVx\x90'






if __name__ == "__main__":

	v1 = 4 # Message (int, 4)
	v2 = 5 # Message (UInt16, 5)

	Integer_0_to_10 = Range (int, 0, 10)

	assert v1 in Integer_0_to_10
	assert v1 not in Range (int, 0, 3)
	assert v1 in ValueList (int, [0, 4, 7])
	assert v1 not in ValueList (int, [0, 3])

	assert "arst" in Length (str, 4)
	assert "arst" not in Length (str, (5, 8))
	assert "arst" not in Length (str, [3, (5, 8)])

	ip = IPv6 (HopLimit = 5, SourceAddress = "fe80::0:0:42")
	ip2 = ip (TrafficClass = 54)

	assert ip     in IPv6 (HopLimit = 5)
	assert ip not in IPv6 (HopLimit = 1)
	assert ip not in IPv6 (HopLimit = 1) (HopLimit = 5)
	assert ip not in IPv6 (HopLimit = 5) (HopLimit = 1)

	udp = UDP (SourcePort = 1025, DestinationPort = 1025, Length = 102)

	ip_udp = pack (ip, udp, "pouet pouet")

	ip_udp_tp = pack (
		IPv6 (HopLimit = 5),
		UDP (SourcePort = 1025, DestinationPort = 1025),
		"pouet pouet"
	)

	ip_udp_tp2 = IPv6 (HopLimit = 5) / UDP (SourcePort = 1025, DestinationPort = 1025) / "pouet pouett"

	ereq = pack (
		IPv6 (SourceAddress = "fe80::ec2a:5eff:fe02:3ec", DestinationAddress = "ff02::1", FlowLabel=0xff00f, TrafficClass=255),
		ICMPv6EchoRequest (Identifier = 0x3a23, SequenceNumber = 1),
		"\0\1\2\3"
	)

	assert ip_udp.flatten() in ip_udp_tp
	assert ip_udp.flatten() not in ip_udp_tp2

	assert Integer_0_to_10.__name__ == "Integer_0_to_10"
	assert ip.__name__ == "ip"
	assert ip2.__name__ == "ip2"
	assert ip_udp.__name__ == "ip_udp"
	assert ip_udp_tp.__name__ == "ip_udp_tp"
	assert ip_udp_tp2.__name__ == "ip_udp_tp2"
	assert IPv6Address.__name__ == "IPv6Address"

	assert ip2[1] == 54
	assert ip2[-8] == 54
	assert ip2["TrafficClass"] == 54

	assert IPv6.__name__ == "IPv6"
	assert IPv6.get_root_variant().__name__ == "IPv6"
	assert UDP.__name__ == "UDP"
	assert ICMPv6.__name__ == "ICMPv6"
	assert ICMPv6Echo.__name__ == "ICMPv6Echo"
	assert ICMPv6EchoRequest.__name__ == "ICMPv6EchoRequest"

	# TODO: some tests on ip_udp_tp & ip_udp

	#print Integer_0_to_10
	#print v1
	#print v2
	#print ip
	#print Range (int, 0, 3)
	#print ValueList (int, [0, 4, 7])
	#print ValueList (int, [0, 3])
	#print Length (str, 4)
	#print Length (str, (5, 8))
	#print Length (str, [3, (5, 8)])
	#print IPv6Message (HopLimit = 5)
	#print IPv6Message (HopLimit = 1)
	#print IPv6Message (HopLimit = 1) (HopLimit = 5)
	#
	UInt8 (54)


	#ip_udp.display()
	#print
	#display (IPv6 (HopLimit = 1) (HopLimit = 5))
	#print
	#display (ereq)

	#result = ip_udp_tp.build_message()

	result = ereq.flatten().build_message()

	result[0].display()
	print (" ".join ([format (v, "02x") for v in result[1]]))
	#TODO: make decode_message accepte bytes
	decoded_ereq = IPv6.decode_message (BinarySlice (result[1]))
	decoded_ereq[0].display()

	Ethernet.decode_message (BinarySlice (b'33\xff463Vu-463\x86\xdd`\x00\x00\x00\x00\x18:\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\xff463\x87\x00\x8e\xe3\x00\x00\x00\x00\xfe\x80\x00\x00\x00\x00\x00\x00Tu-\xff\xfe463'))[0].display()

	val, bin = pack(
		Ethernet (SourceAddress = "00:01:02:03:04:05"),
		IPv6 (SourceAddress = "fe80::42", HopLimit=255),
		ICMPv6NeighborAdvertisement (Flags = 0x6000, TargetAddress = "fe80::42",
		Options = [ICMPv6TLLOption (LinkLayerAddress = "00:01:02:03:04:05")]),
	).flatten().build_message()
	val.display()
	print (bin)

# idees
# template parametre
#
#def DHCP_Req (source):
#	return pack (
#		L2Hdr (src = source.L2),
#		IPv6 (src = source.IP),
#		UDP (sport = 547, dport = 546).
#		DHCPv6Request (opt = [DHCPv6ClientId (id = source.id)])
#	)
#
#
#link1.send (
#	DHCP_Req (source = TN1) (
#		ipv6_source = "blahblah",
#		dhcpv6_opt_add = id
#	),
#
#	DHCP_Req (source = TN1) ({	# crade
#		IPv6: {"source": blahblah},
#		DHCPv6: {"opt_add": id}
#	})
#)
#
"""



from ..common import *
from ...core.templates import Range

class TD_COAP_BLOCK_04(CoAPTestCase):
    """
---
TD_COAP_BLOCK_04:
    obj: Handle POST blockwise transfer for large resource
    cfg: CoAP_CFG_BASIC
    ref: '[BLOCK] 2.2, 2.3, 2.5'

    pre:
        - Client supports Block1 transfers
        - Server supports Block1 transfers
        - Server accepts creation of new resources on /large-create

    seq:
    -   s: 'Client is requested to create a new resource /large-create on
        Server'

    -   c:
        - 'Client sends a POST request containing Block1 option
        indicating:'
        -   - NUM = 0
            - M = 1
            - SZX = the desired block size
            - Payload size is 2SZX+4 bytes

    -   c:
        - 'Server sends 2.31 (Continue) response containing Block1 option indicating:'
        -   - NUM = 0
            - M = 1 (atomic)
            - SZX is less or equal to the SZX at step 2

    -   c:
        - 'Client sends further requests containing Block1 option indicating:'
        -   - 'NUM = i where “i” is the block number of the current
            block. If the server decreased the SZX parameter in
            step 3, then the client should adapt the block size
            accordingly and may resume the transfer from block
            id 2size_in_step_2-size_in_step_3 instead of block 1)'
            - M = 1 (more)
            - SZX is the SZX at step 3
            - "Payload size is 2SZX+4 bytes"

    -   c:
        - 'Server sends 2.01 (Created) response containing
        Block1 option indicating:'
        -   - NUM = i where “i” is the block number used at step 4
            - M = 1 (atomic)
            - SZX is the SZX at step 3

    -   c:
        - 'Client send PUT request containing the last block and
        indicating:'
        -   - NUM = n where “n” is the last block number;
            - M = 0 (final)
            - SZX is the SZX at step 3
            - Payload size is lesser or equal to 2SZX+4

    -   c:
        - 'Server sends 2.01 (Created) response containing Block1
        option indicating:'
        -   - NUM = n where “n” is the block number used at step 6
            - M = 0
            - SZX is the SZX at step 3
        - 'and Location-Path options, e.g. if the new location
            is "/large-create/PS":'
        -   - First option value must contain “large-create”
            - Second option value is a (single) path segment chosen by the server (PS)
            - none of the Location-Path options contain a ‘/’

    -   v: Client displays the response

    -   v: 'Server indicates presence of the complete new resource,
            e.g., /large-create/PS'

    -   c:
        - 'Client sends GET request to new location (e.g., "/large-create/PS",
            i.e., using Uri-Path options simply copied from the Location-Path
            of step 7)'

    -   c: 'Server sends 2.05 (Content) response with representation
            of created resource, potentially making use of the Block2 protocol'

    -   v: Client indicates the value of the newly created resource

    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [
            CoAP(type='con', code='post', opt=Opt(CoAPOptionUriPath("large-create"))),
        ]

    def run (self):
        # Step 2 - initial request from the client
        self.match("client", CoAP (code = "post",
                         opt = self.uri ("/large-create", CoAPOptionBlock1 (num=0, m=1))))
        client_szx  = self.coap["opt"][CoAPOptionBlock1]["szx"]
        client_size = 2**(client_szx+4)
        self.match("client", CoAP (pl = Length (bytes, client_size)), "fail")

        self.next_skip_ack()

        # Step 3 - response from the server possibly w/ block size negociation
        if not self.match("server", CoAP (code = 2.31,
                         opt = Opt (CoAPOptionBlock1 (num=0,m=1, szx=Range(int, 0, client_szx)))
                )):
            return
        server_szx  = self.coap["opt"][CoAPOptionBlock1]["szx"]
        server_size = 2**(server_szx+4)

        self.next_skip_ack()

        # next block should have num 1
        counter = itertools.count (1)

        if server_szx != client_szx:
            # client & server do not agree on the block size

            # server requested a smaller block size, num may be bigger
            if (self.match("client", CoAP (opt = Opt (CoAPOptionBlock1()))), None):
                num = self.coap["opt"][CoAPOptionBlock1]
                if num > 1:
                    counter = itertools.count (num)

        for i in counter:

            # Step 4 - send next block from client
            self.match("client", CoAP (code = "post",
                             opt = Opt (CoAPOptionBlock1(num=i, szx=server_szx))
                    ))

            # Step 4 - more blocks.
            last = not self.match("client", CoAP (    opt = Opt (CoAPOptionBlock1(m=1)),
                                    pl = Length (bytes, server_size)
                    ), None)
            if last:
                # Step 6 - last block
                self.match("client", CoAP (    opt = Opt (CoAPOptionBlock1 (m=0)),
                                    pl = Length (bytes, (1, server_size))
                        ), "fail")

            self.next_skip_ack()

            # Step 5/Step 7 - ack from the server
            if not self.match("server", CoAP (code = Any (CoAPCode (2.01), 2.31),
                             opt = Opt (CoAPOptionBlock1(num=i, szx=server_szx))
                    )):
                raise self.Stop()

            if last:
                # Step 7
                if not self.match("server",CoAP(opt=self.uri ("/large-create",
                        CoAPOptionBlock1 (m=0),
                        CoAPOptionLocationPath(),
                        CoAPOptionLocationPath()
                    )),"fail"):
                    raise self.Stop()
                break
            self.next_skip_ack (optional = True)
        # LocationPath value
        for o in self.coap["opt"]:
            if CoAPOptionLocationPath(dlt=0).match(o):
                valPath=o["val"]
        self.next(optional = True)
        # Step 10
        if self.match("client",CoAP(code="get",opt=self.uri ("/large-create/"+valPath,CoAPOptionUriPath(),CoAPOptionUriPath(valPath))),None):
            self.next_skip_ack ()
            # Step 11
            self.match("server",CoAP(code=2.05,opt=Opt(CoAPOptionBlock2())))
        raise self.Stop()



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_12(CoAPTestcase):
    """Identifier:
TD_COAP_CORE_12
Objective:
Perform GET transaction using empty Token (CON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] 2.2 ,5.8.1, 5.10.1

Pre-test
conditions:
•	Server offers the resource /test with resource content is not empty that handles GET with an arbitrary payload

Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a confirmable GET request using zero-length Token to server’s resource

2
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Zero-Length Token ➔ CTOK
•	Uri-Path option "test"

3
Check
Server sends response containing:
•	Code = 69 (2.05 content)
•	Message ID = CMID, Token = CTOK
•	Not empty Payload
•	Content format option

4
Verify
Client displays the response
"""

    def run(self):
        self.match_coap("client", CoAP(type="con", code="get",
                                       tok=b"",
                                       opt=self.uri("/test")))

        self.next_skip_ack()

        if self.match_coap("server", CoAP(code=2.05,
                                          pl=Not(b""),
                                          opt=Opt(CoAPOptionContentFormat()),
                                          )):
            self.match_coap("server", CoAP(tok=b""), "fail")



import unittest
from os import path

from ttproto.core.dissector import Frame, Capture, ReaderError


class CaptureTestCase(unittest.TestCase):
    """
    Test class for the capture tool
    """

    # #################### Tests parameters #########################

    # File path
    TEST_FILE_DIR = 'tests/test_dumps'
    PCAP_FILE = path.join(TEST_FILE_DIR, 'coap', 'CoAP_plus_random_UDP_messages.pcap')

    # pcaps that MUST throw exceptions
    WRONG_TEST_FILE_DIR_NAME = 'others'
    EMPTY_PCAP_FILE = path.join(TEST_FILE_DIR, WRONG_TEST_FILE_DIR_NAME, 'empty_pcap.pcap')
    NOT_A_PCAP_FILE = path.join(TEST_FILE_DIR, WRONG_TEST_FILE_DIR_NAME, 'not_a_pcap_file.dia')

    CAPTURE_LENGTH = 5

    # #################### Init and deinit functions #########################
    def setUp(self):
        """
            Initialize the Capture instance
        """
        self.capture = Capture(self.PCAP_FILE)

    # #################### Tests functions #########################

    # ##### filename
    def test_get_filename(self):
        self.assertEqual(self.capture.filename, self.PCAP_FILE)

    def test_set_filename(self):
        with self.assertRaises(AttributeError):
            self.capture.filename = 'New filename'

    # ##### frames
    def test_get_frames(self):
        self.assertEqual(
            len(self.capture.frames) + len(self.capture.malformed),
            self.CAPTURE_LENGTH
        )
        for frame in self.capture.frames:
            self.assertIsInstance(frame, Frame)

    def test_set_frames(self):
        with self.assertRaises(AttributeError):
            self.capture.frames = []

    # ##### frames
    def test_get_malformed(self):
        self.assertEqual(
            len(self.capture.malformed) + len(self.capture.frames),
            self.CAPTURE_LENGTH
        )
        for malformed_frame in self.capture.malformed:
            self.assertIsInstance(malformed_frame, Frame)
            self.assertIsNotNone(malformed_frame['error'])

    def test_set_malformed(self):
        with self.assertRaises(AttributeError):
            self.capture.malformed = []


# #################### Main run the tests #########################
if __name__ == '__main__':
    unittest.main()



from ..common import *


class TD_LWM2M_1_INT_241(CoAPTestCase):
    """
---
testcase_id: TD_LWM2M_1.0_INT_241
uri : http://openmobilealliance.org/iot/lightweight-m2m-lwm2m
configuration: LWM2M_CFG_01
objective: 
  - Test that the Device can be remotely rebbot via the Device Object (ID:3)
pre_conditions: 
  - The Client supports the minimum Configuration C.1.
  - Device is switched on
  - Client is registered with the Server. (Client might be in a state that requires a reboot)
references: 'OMA-ETS-LightweightM2M-V1_0_1-20170926-A'
sequence:
  - step_id: 'TD_LWM2M_1.0_INT_241_step_01'
    type: stimuli
    node : lwm2m_server
    description:
      - 'LwM2M server performs Execute (CoAP POST) operation on Reboot Resource of Device Object (ID:3) Instance'
      - - Type = 0 (CON)
        - Code = 2 (POST)

  - step_id: 'TD_LWM2M_1.0_INT_241_step_02'
    type: check
    description:
      - 'The request sent by the server contains'
      - - Type=0 and Code=2
        - Uri-Path option= 3/0/4

  - step_id: 'TD_LWM2M_1.0_INT_241_step_03'
    type: check
    description:
        - 'LwM2M client sends response containing'
        - - Code = 2.04 (Changed)

  - step_id: 'TD_LWM2M_1.0_INT_241_step_04'
    type: verify
    node: LwM2M Server
    description:
        - 'Device reboots successfully and re-registers at the server again.'

  - step_id: 'TD_LWM2M_1.0_INT_241_step_05'
    type: stimuli 
    node: lwm2m_client
    description:
      - 'LwM2M client sends a Registration message (CoAP POST) to LwM2M server'
      - - Type = 0 (CON)
        - Code = 2 (POST)
        - Content-format option=application/link-format
        - Non-empty Payload

  - step_id: 'TD_LWM2M_1.0_INT_241_step_06'
    type: check
    description:
      - 'Sent POST request contains'
      - - Type=0 and Code=2
        - Non-empty payload= A Media-Type payload where each Object is described as a Link according to content-format option. 
        - content-format=application/link-format
        - Uri-Path option = /rd
        - Uri-Query option = 'ep={endPointClientName}'
        - Uri-Query option = 'lt={lifetime}'
        - Uri-Query option = 'sms={MSISDN}'
        - Uri-Query option = 'lwm2m={version}'
        - Uri-Query option = 'b={binding}'

  - step_id: 'TD_LWM2M_1.0_INT_241_step_07'
    type: check
    description:
      - 'LwM2M server sends response containing'
      - - Code = 2.01 (Created)
        - content-format=application/link-format
        - LocationPath option = the path to use for updating or deleting the registration
        - Non-empty payload

  - step_id: 'TD_LWM2M_1.0_INT_241_step_08'
    type: verify
    node: lwm2m_client
    description:
      - 'LwM2M client indicates successful operation'

    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [CoAP(type='con', code='post')]

    def run(self):
        self.match('server', CoAP(type='con', code='post', opt=self.uri('/3/0/4')), 'fail')
        self.match('server', CoAP(pl=(b'')), 'fail')

        self.next()

        self.match('client', CoAP(code=Any(65, 68)), 'fail')

        self.next()

        self.match('client', CoAP(type='con', code='post', opt=self.uri('/rd')), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionContentFormat('40'))), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionAccept('40'))), 'fail')
        opts = list(filter((lambda o: isinstance(o, CoAPOptionUriQuery)), self.frame.coap["opt"]))
        if len(opts) < 1:
            self.setverdict("inconclusive", "expect at least one UriQuery option")
        self.match('client', CoAP(opt=Opt(CoAPOptionContentFormat('40'))), 'fail')

        self.next()

        self.match('server', CoAP(code=2.01, pl=Not(b'')), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionContentFormat('40'))), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionLocationPath())), 'fail')



#!/usr/bin/env python3
#
#  (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from ttproto import TMPDIR
from ttproto.utils.pure_pcapy import *
from ttproto.core.typecheck import *

# TODO extend to other types of profiles which involve sth different that removing the first X bytes of each frame
# TODO filter first X layers of pcap
# TODO protocol filter: ex: filter anything that's not IEE802.15.4? dissect()  already does this


# #################### Filter Functions #########################

@typecheck
def remove_first_bytes(number_of_bytes: int, new_snaplen: int, new_network, pcap_filename: str,
                       new_pcap_filename: str = TMPDIR + "/temp.pcap"):
    reader = open_offline(pcap_filename)
    dumper = Dumper(new_pcap_filename, new_snaplen, new_network)

    count = 0
    stop_iter = False
    new_header, new_data = reader.next()

    while not stop_iter:
        try:
            count += 1
            new_header.incl_len = new_header.incl_len - number_of_bytes
            new_header.orig_len = new_header.incl_len
            new_data = new_data[number_of_bytes:]  # take from jumplength bytes to the end of the data
            dumper.dump(new_header, new_data)
            new_header, new_data = reader.next()
            if not new_header:
                stop_iter = True
                print('The filtering has reached the end of the file :D, packet count %d' % count)
        except PcapError as e:
            print(e)
            break

@typecheck
def remove_first_frames(number_of_frames_to_skip: int,
                        pcap_filename: str,
                        new_pcap_filename: str = TMPDIR + "/temp.pcap"):

    reader = open_offline(pcap_filename)
    dumper = Dumper(new_pcap_filename, reader.snaplen, reader.network)

    count = 0
    stop_iter = False

    for _ in range(0, number_of_frames_to_skip):
        reader.next()

    while not stop_iter:
        try:
            new_header, new_data = reader.next()
            count += 1
            new_data = new_data
            dumper.dump(new_header, new_data)
            if not new_header:
                print('The filtering has reached the end of the file :D, packet count %d' % count)
                break
        except PcapError as e:
            print(e)
            break


# #################### Filter Profiles #########################

@typecheck
def openwsn_profile_filter(pcap_filename: str, new_pcap_filename: str = TMPDIR + '/temp.pcap') -> str:
    """
    This filter is usded to filter the extra layers added by openwsn openvisualizer when sniffing ieee802.15.4
    and forwarding to the tun/tap interface
    For openWSN/openvisualizer captures we need to ignore the first 5*16 bytes (as it always generates raw:ipv6:udp:zep:wpan)

    :param pcap_filename:
    :param new_pcap_filename:
    :return: filename of new pcap file
    """
    JUMP_LENGTH = 16 * 5  # en bytes
    remove_first_bytes(JUMP_LENGTH, 200, DLT_IEEE802_15_4, pcap_filename, new_pcap_filename)
    return new_pcap_filename


@typecheck
def finterop_tun_profile_filter(pcap_filename: str, new_pcap_filename: str = TMPDIR + '/temp.pcap') -> str:
    """
    For rewriting link type on f-interop captures
    :param pcap_filename:
    :param new_pcap_filename:
    :return: filename of new pcap file
    """
    JUMP_LENGTH = 0  # en bytes
    remove_first_bytes(JUMP_LENGTH, 200, DLT_RAW, pcap_filename, new_pcap_filename)
    return new_pcap_filename



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from ttproto.core.data import *
from ttproto.core.typecheck import *
from ttproto.core import port, clock
from ttproto.utils import pure_pcapy


class PcapPort(port.RawMessagePort):
    @typecheck
    def __init__(self, file: str, decode_type: optional(is_type) = None, clock_: optional(clock.Clock) = None,
                 endpoint: optional(port.BaseMessagePort) = None):

        self.__reader = pure_pcapy.Reader(open(file, "rb"))
        self.__clock = clock_ if clock_ else clock.Clock.get_instance()

        if not decode_type:
            try:
                decode_type = _map_link_type[self.__reader.datalink()]
            except KeyError:
                decode_type = bytes

        port.RawMessagePort.__init__(self, decode_type, endpoint)

        self.__schedule_next()

    def __schedule_next(self):

        h, self.__next_packet = self.__reader.next()

        if h:
            # timestamp
            ts = h.getts()
            ts = ts[0] + ts[1] * 0.000001

            # schedule
            self.__clock.schedule_event_absolute(ts, self.__callback)

    def __callback(self):
        self._forward(self.__next_packet)

        self.__schedule_next()

    def enqueue(self, msg):
        raise Exception("PcapPort is receive-only")



from ..common import *


class TD_M2M_NH_12(CoAPTestCase):
    """
---
TD_M2M_NH_12:
    cfg: M2M_CFG_01
    obj: AE updates attribute in application resource via <container> Update Request
    pre: AE has created an Application Entity resource <AE> on Registrar CSE and
         AE has created a container resource <container> on Registrar CSE
    ref: 'TS-0001, clause 10.2.4.3 and TS-0004, clause 7.3.5.2.3'
    seq:
    -   s:
        - 'AE is requested to send a PUT request to update the lifetime of <container> resource with:'
        -   - Type = 0 (CON)
            - Code = 3 (PUT)
            - Non-empty payload
            - Content-format option
    -   c:
        - 'The request sent by the client contains:'
        -   - Type=0 and Code=3
            - oneM2M-FR option=AE-ID
            - oneM2M-RQI=token-string (CRQI)
            - contentFormat option
            - Non-Empty payload      
    -   v: Check if possible that the <container> resource has been updated in registrar CSE. 
    -   c:
        - 'Registrar CSE sends response containing:'
        -   - Type = 0 (CON)
            - Code = 2.04 (UPDATED)
            - Token = CTOK
            - Content-format option
            - non-empty Payload
            - oneM2M-RSC option = 2004
            - oneM2M-RQI=CRQI
    -   v: AE indicates successful operation
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """

        return [CoAP(type='con', code='put')]

    def run(self):
        self.match('client', CoAP(type="con", code="put", pl=Not(b'')), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionContentFormat())), "fail")
        self.match('client', CoAP(opt=Opt(CoAPOptionOneM2MFrom())), 'fail')
        if self.match('client', CoAP(opt=Opt(CoAPOptionOneM2MRequestIdentifier())), 'fail'):

            OPTS = self.coap['opt']
            RI = OPTS[CoAPOptionOneM2MRequestIdentifier]
            RIVAL = RI[2]

        
            self.next()

            self.match("server", CoAP(type='con', code=Any(65, 68), pl=Not(b'')), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionContentFormat())), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionOneM2MResponseStatusCode('2004'))), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionOneM2MRequestIdentifier(RIVAL))), 'fail')



#!/usr/bin/env python3

#from ttproto.ts_coap.proto_specific import CoAPTestcase
# from ttproto.ts_coap.proto_templates import *
from ttproto.ts_coap.testcases.td_coap_core_20 import TD_COAP_CORE_20

class TD_COAP_CORE_26 (TD_COAP_CORE_20):
	"""Identifier:
TD_COAP_CORE_26
Objective:
Perform GET transaction containing the Accept option (CON mode
Configuration:
CoAP_CFG_03
References:
[1] clause 5.8.1,5.10.5,5.10.4, 8.2.2,8.2.1,10.2.2,11.2

Pre-test
conditions:
•	Proxy is configured as a reverse-proxy for the server
•	Proxy’s cache is cleared
•	Server should provide a resource /multi-format which exists in two formats:
-	text/plain;charset=utf-8
-	application/xml

Test Sequence:
Step
Type
Description
Part A: client requests text format


1
Stimulus
Client is requested to send a confirmable GET request to
proxy

2
Check
Proxy receives the request from client  & forwards it to
server’s resource


3
Check
Forwarded request must contain:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Option: type = Accept, value = 0 (text/plain;charset=utf-8)

4
Check
Server sends response containing:
•	Code = 69 (2.05 content)
•	Option: type = Content-Format, value = 0
(text/plain;charset=utf-8)
•	Payload = Content of the requested resource in
text/plain;charset=utf-8 format

5
Check
Proxy forwards the response to client

6
Verify
Client receives  & displays the response

7
Check
Response contains:
•	Code = 69 (2.05 content)
•	Option: type = Content-Format, value = 0
(text/plain;charset=utf-8)
•	Payload = Content of the requested resource in
text/plain;charset=utf-8 format

Part B: client requests xml format

8
Stimulus
Client is requested to send a confirmable GET request to
Proxy

9
Check
Proxy forwards the request to server

10
Check
Sent request must contain:
•	Type = 0 (CON)
•	Code = 1 (GET)
Option: type = Accept, value = 41 (application/xml)

11
Check
Server sends response containing:
•	Code = 69 (2.05 content)
•	Option: type = Content-Format, value = 41
(application/xml)
Payload = Content of the requested resource in
application/xml format

12
Check
Proxy forwards the response to client

13
Verify
Client receives & displays the response

14
Check
Client displays the response received:
•	Code = 69 (2.05 content)
•	Option: type = Content-Format, value = 41
(application/xml)
Payload = Content of the requested resource in
application/xml format
	"""
	reverse_proxy = True




#!/usr/bin/env python3

from ttproto.core import typecheck3000
from ttproto.core.typecheck3000 import *

'''
TODO: hack typeckeck3000 to display the line number locating the function
that threw the exception

TODO: implémenter un typecheck_inherit
'''

#TODO: implement it
this_class = anything


__all__ = list(typecheck3000.__all__)
__all__.append("this_class")



from ttproto.core.typecheck3000 import iterable	# missing symbol (TODO: bug report)
__all__.append ("iterable")

#typecheck3000.disable()




from ..common import *


class TD_COAP_OBS_06(CoAPTestCase):
    """
---
TD_COAP_OBS_06:
    obj: Server detection of deregistration (explicit RST)
    cfg: CoAP_CFG_BASIC
    ref: '[OBSERVE] 4.2.2'
    pre:
        - Client supports Observe option
        - Server supports Observe option
        - Server offers an observable resource /obs which changes periodically (e.g. every 5s) which produces confirmable notifications

    seq:
    -   s:
        - 'Client is requested to send to the server a confirmable GET request with'
        -  - observe option for resource /obs

    -   c:
        - 'The request sent by client contains:'
        - Type = 0 (CON)
        - Code = 1 (GET)
        - Token value  = a value generated by the client
        - Observe option = empty

    -   c:
        - 'Server sends the response containing:'
        - Type = 2 (ACK)
        - Code = 2.05 (Content)
        - Content-format of the resource /obs
        - Token value = same as one found in the step 2
        - Observe option with a sequence number
    -   s:
        - 'Server sends a notification containing:'
        - Type = 0 (CON)
        - Code = 2.05 (Content)
        - Content-format = same as one found in the step 3
        - Token value = same as one found in the step 2
        - Observe option indicating increasing values

    - v: Client displays the received information

    -   c:
        - 'Client sends an ACK'
    -   s:
        - 'Client is rebooted'

    -   c:  'Server is still sending notifications for the request in step 2 as in step 4'
    -   v: 'Client discards response and does not display information'

    -   c: 'Client sends RST to Server'

    -   v: Server does not send further response

    -   v: Client does not display further received information

    """
    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [
            CoAP(type='con', code='get', opt=Opt(CoAPOptionObserve(0),CoAPOptionUriPath("obs"))),
        ]

    def run (self):
        request = CoAP(type="con", code="get", opt=self.uri("/obs", CoAPOptionObserve (0)))
        response= CoAP(type="ack", code=2.05, opt=Opt(CoAPOptionObserve(), CoAPOptionContentFormat()))
        notif   = CoAP(type="con", code=2.05, opt=Opt(CoAPOptionObserve(), CoAPOptionContentFormat()))

        # Step 2
        self.match("client", request)

        token = self.coap["tok"]
        uri = self.coap.get_uri()

        self.next()

        # Step 3
        if not self.match("server", response):
            raise self.Stop()
        self.match("server", CoAP(tok = token), "fail")

        content_format = self.coap["opt"][CoAPOptionContentFormat]["val"]
        index          = self.coap["opt"][CoAPOptionObserve]["val"]

        self.next()

        # A list of message CMID that did not receive ACK from the client.
        CMID_without_ACK = set()
        if self.match("server", CoAP(type = "con", code=2.05, tok=token), None):
            CMID_without_ACK.add(self.coap["mid"])

        continue_loop = True
        rst_sent_by_client = False
        # Step 4-6
        while self.match("server", CoAP(type = "con", code=2.05, tok=token), None) and continue_loop:
            CMID_without_ACK.add(self.coap["mid"])
            self.match("server", CoAP(tok = token), "fail")
            self.match("server", CoAP(opt = Opt(CoAPOptionContentFormat (content_format))), "fail")

            # increasing observe value
            new_index = self.coap["opt"][CoAPOptionObserve]["val"]
            self.set_verdict(("pass" if new_index > index else "fail"),
                "value of observe option must be increasing")
            index = new_index
            last_frame = self._frame # see comment in the else: block
            self.next(True)
            if self.match("client", CoAP(type = "ack"), None):
                CMID_without_ACK.remove(self.coap["mid"])
                self.next()
            elif self.match("client", CoAP(type = "rst", code=0), None) :
                # Step 10
                rst_sent_by_client = True
                self.next(True)
            else:
                # Avoid infinite loop in the scenario where the client do not
                # send an RST. We could just put the above elif block outside
                # the loop as another solution, but it may make harder to do
                # more verification (e.g for dedug purpose or additional
                # verifications) in the event where it would have more frame
                # after the client sent an RST.
                if last_frame == self._frame:
                    continue_loop = False

        if len(CMID_without_ACK) == 0 and not rst_sent_by_client:
                        self.set_verdict("inconclusive", "There is no \
unacknowledged notification nor RST.")
        elif len(CMID_without_ACK) == 0 and rst_sent_by_client:
            self.set_verdict("inconclusive", "The client did send an RST \
before rebooting. Hence we do not have any unacknowledged notification. We\
cannot make any conclusion.")
        elif len(CMID_without_ACK) < 2 and not rst_sent_by_client:
            # STEP 8
            self.set_verdict("fail", "STEP 8: Server did not continue \
to send notifications as expected. It should continue sending some notification \
even after the first unacknowledged one UNTIL it received a RST \
or if enough time elapsed.\
In other words, while the client did not send ACK nor RST we must see \
unacknowledged notification(s) in the list of frames to obtain a pass verdict.")
        # Step 10 -- Check : Client sends RST to Server.
        # At this point STEP 8 is ok.
        elif rst_sent_by_client:
            self.set_verdict("pass", "STEP 8 : Server is still sending \
notifications for the request in step 2 as in step 4")
            self.set_verdict("pass", "STEP 10: RST has been sent by the client.")
        else:
            self.set_verdict("pass", "STEP 8 : Server is still sending \
notifications for the request in step 2 as in step 4")
            self.set_verdict("fail", "STEP 10: RST has not been sent by the client.")



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_09 (CoAPTestcase):
    """Identifier:
TD_COAP_CORE_09
Objective:
Perform GET transaction with separate response (CON mode, no piggyback)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] clause  5.8.1,5.2.2

Pre-test
conditions:
•	 Server offers a resource /separate which is not served immediately and which therefore is not acknowledged in a piggybacked way.
cannot be acknowledged in a piggybacked way.

Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a confirmable GET request to server’s resource

2
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Client-generated Message ID (➔ CMID)
•	Uri-Path option "separate"
•	Client-generated Token (➔ CTOK)


3
Check
Server sends response containing:
•	Type = 2 (ACK)
•	Code = 0
•	Message ID = CMID
•	empty Payload

4
Check
Server sends response containing:
•	Type  = 0 (CON)
•	Code = 69 (2.05 content)
•	Server-generated Message ID (➔ SMID)
•	Token = CTOK
•	Content-format option
•	Non-empty Payload

5
Check
Client sends response containing:
•	Type = 2 (ACK)
•	Code = 0
•	Message ID = SMID
•	empty Payload

6
Verify
Client displays the response
Note: Steps 3 and 4 may occur out-of-order
"""
    def run (self):
        self.match_coap ("client", CoAP (type="con", code = "get",
                        opt = self.uri ("/separate")))
        CMID = self.frame.coap["mid"]
        CTOK = self.frame.coap["tok"]

        self.next()

        #FIXME: may be out-of-order
        if not self.match_coap	("server", CoAP (type="ack", code=0, mid=CMID, pl = b"")):
            raise self.Stop()

        self.next()

         # FIXME: this is in a different conversation
        self.match_coap ("server", CoAP (type="con", code=2.05))
        self.match_coap ("server", CoAP (
                        tok=CTOK,
                        pl = Not (b''),
                        opt= Opt(CoAPOptionContentFormat())
                ), "fail")
        SMID = self.frame.coap["mid"]

        self.next()

        self.match_coap ("client", CoAP (type="ack", code=0, mid=SMID,pl=b""))





#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re
import logging


from    ttproto.core.exceptions     import Error, DecodeError
from    ttproto.core.data       import *
from    ttproto.core.typecheck      import *
from    ttproto.core.packet     import *
from    ttproto.core.union      import *
from    ttproto.core.lib.inet.basics    import *
from    ttproto.core.lib.inet.meta  import FixedLengthBytesClass
from    ttproto.core.lib.inet.sixlowpan import SixLowpan
from    ttproto.core.lib.inet.sixlowpan_hc import SixLowpanIPHC


#logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger('[802.15.4 - CoDec]')

__all__ = [
    'Ieee802154ShortAddress',
    'Ieee802154Address',
    'Ieee802154',
    #'Ieee802154Ack',
]


_address_mode_descriptions = {
    0:  "not present",
    1:  "reserved",
    2:  "short",
    3:  "extended",
}


class _Reversed (PacketValue.Tag):
    def build_message (self, value, ctx):
        v, b = value.build_message()

        assert isinstance (b, bytes)  # unaligned bins not supported

        return v, bytes (reversed (b))

    def decode_message (self, type_, bin_slice, ctx):
        v, result_slice = type_.decode_message (bin_slice)

        length = result_slice.get_left() - bin_slice.get_left()

        assert result_slice.same_buffer_as (bin_slice)
        assert length % 8 == 0

        v, tmp_slice = type_.decode_message (BinarySlice (bytes (reversed (bin_slice.bit_slice (0, length)))))

        assert not tmp_slice # must be fully decoded

        return v, result_slice

class HexUInt16 (UInt16):
    def __new__ (cls, value):
        if isinstance (value, bytes):
            assert len (value) == 2
            return super().__new__ (cls, value[0]*256 + value[1])
        else:
            return super().__new__ (cls, value)

    def __str__ (self):
        return hex(self)

class Ieee802154ShortAddress (metaclass = FixedLengthBytesClass (2)):
    __reg = re.compile ("([0-9A-F]{2})[-:.]?([0-9A-F]{2})\\Z", re.I)

    def __new__ (cls, value = None):
        if isinstance (value, str):
            mo = re.match (cls.__reg, value)
            if not mo:
                raise Error("malformed IEEE 802.15.4 short address")

            value = bytes (int (v, 16) for v in mo.groups())

        return super().__new__ (cls, value)


class Ieee802154Address (
    metaclass = UnionClass,
    types     = (Ieee802154ShortAddress, Eui64Address),
):
    def __new__ (cls, value):
        if (isinstance (value, str) and len(value) > 5) or (isinstance (value, bytes) and len(value) > 2):
            return Eui64Address (value)
        else:
            if isinstance (value, int):
                log.warning("***warning: deprecated*** Ieee802154 short addresses are now coded as bytes")
                value = bytes ((value // 256, value % 256))

            return Ieee802154ShortAddress (value)



# TODO re-reimplement code/decode of ACK,DATA, BEACON and MAC control frames  with variants!

class Ieee802154 (
    metaclass = PacketClass,
    fields    = [
        ("FrameType",           "type", UInt3,      1),
        ("SecurityEnabled",     "se",   bool,       False),
        ("FramePending",        "fp",   bool,       False),
        ("AcknowlegeRequest",       "ar",   bool,       False),
        ("IntraPan",            "ip",   bool,       False),
        ("Reserved",            "rsv1", UInt3,      0),
        ("DestinationAddressingMode",   "dam",  UInt2,      0),
        ("FrameVersion",        "ver",  UInt2,      0),
        ("SourceAddressingMode",    "sam",  UInt2,      0),

        ("SequenceNumber",      "seq",  UInt8,      0),
        ("DestinationPanId",        "dpid", Optional (HexUInt16),       _Reversed(Omit())),
        ("DestinationAddress",      "dst",  Optional (Ieee802154Address),   _Reversed(Omit())),
        ("SourcePanId",         "spid", Optional (HexUInt16),       _Reversed(Omit())),
        ("SourceAddress",       "src",  Optional (Ieee802154Address),   _Reversed(Omit())),
        ("Payload",         "pl",   Value,      ""),
        ('FCS',     'fcs',  Optional ( HexUInt16), _Reversed(Omit()))
    ],

    #id = 1, # Data frame by default

    descriptions = {
        "type": {
            0:  "Beacon Frame",
            1:  "Data Frame",
            2:  "Ack Frame",
            3:  "MAC Command Frame",
        },
        "sam":  _address_mode_descriptions,
        "dam":  _address_mode_descriptions,
        "ver": {
            0:  "IEEE 802.15.4-2003",
            1:  "IEEE 802.15.4-2006",
            2:  "future use",
            3:  "future use",
        },
    }):

    def describe (self, desc):
        desc.hw_src = self["src"]
        desc.hw_dst = self["dst"]
        if not self.describe_payload (desc):
            desc.info = "IEEE 802.15.4"

        return True

    def _build_message (self):
        values = self.DataList (self._fill_default_values())

        # dst addr mode
        if self["dam"] is None and not isinstance (values["dst"], Omit):
            values["dam"] = 3 if isinstance (values["dst"], Eui64Address) else 2
        # dst pan id
        if self["dpid"] is None and values["dam"]:
            # set it to 0x0000 by default
            values["dpid"] = 0

        # src addr mode
        if self["sam"] is None and not isinstance (values["src"], Omit):
            values["sam"] = 3 if isinstance (values["src"], Eui64Address) else 2

        # intra_pan
        if not isinstance (values["sam"], Omit) and isinstance (values["spid"], Omit):
            values["ip"] = store_data (True, self.get_field ("ip").type)

        # enter a ieee_addresses context before encoding the fields so that the current
        # hw source & destination addresses are known to the upper layers
        # (needed by 6lowpan-hc)
        with SixLowpanIPHC.encapsulating_iid_context (values["src"], values["dst"]):
            values, bins = zip(*(f.tag.build_message(v, None) for f,v in zip (self.fields(), values)))

        # pack the frame control field
        seq_id = self.get_field_id ("seq")
        b = concatenate (reversed (bins[:seq_id]))
        # reverse the bytes order
        fc = bytes ((b[1], b[0]))

        return type(self) (*values), concatenate ((fc,) + bins[seq_id:])

    @classmethod
    def _decode_message (cls, bin_slice):
        #log.debug('[header] Starting to decode 802.15.4 header')
        seq_id = cls.get_field_id ("seq")

        def reversed_slice (sl):
            assert sl.get_bit_length() % 8 == 0
            return BinarySlice(bytes(reversed(sl.raw())))

        def decode_field (i, sl, t = None):
            f = cls.get_field (i)
            v, sl = f.tag.decode_message (t if t else f.type, sl, None)
            values.append (v)
            #log.debug('[header] [field] : ' + str(f)  + " || value : " +str(v))
            return sl


        def decode_FCS(bin_slice):
            bin_slice = decode_field ('fcs', bin_slice, HexUInt16)
            #log.debug(len(bin_slice))
            #log.debug(str(bin_slice))
            #log.debug(str(cls.get_field))
            return bin_slice

        # decode the Frame Control Field
        fc_slice = reversed_slice (bin_slice[:2])

        values = []
        for i in range(seq_id-1,-1,-1):
            fc_slice = decode_field (i, fc_slice)
            #log.debug('Decoding FC: ' + str(values))
        values = cls.List (reversed (values))

        #log.debug('Decoded FC: '+ str(values))

        assert not fc_slice

        bin_slice = bin_slice[2:]

        # sequence number
        bin_slice = decode_field (seq_id, bin_slice)
        #log.info('[header] [field] Decoded seq_id: ' + str(values["seq"]))

        if values["type"]==2:
            #log.info("[header] [field] The 802.15.4 message is an ACK")
            # let's make an educated guess and assume there's FCS if left bin_slice == 2
            if len(bin_slice) == 2:
                bin_slice = decode_FCS(bin_slice)
            assert len(bin_slice)==0
            return cls(*values), bin_slice

        # addresses
        sam = values ["sam"]
        dam = values ["dam"]
        intra = values ["ip"]

        am_type = (Omit, Omit, Ieee802154ShortAddress, Eui64Address)
        am_length = (0, 0, 2, 8)

        # dst pan-id
        bin_slice = decode_field ("dpid", bin_slice, None if dam else Omit)
        # dst address
        bin_slice = decode_field ("dst", bin_slice, am_type[dam])

        # src pan-id
        bin_slice = decode_field ("spid", bin_slice, Omit if (intra or not sam) else None)

        # src address
        #log.debug(str(am_type[sam]))
        bin_slice = decode_field ("src", bin_slice, am_type[sam])

        #log.debug('[header] finished processing header. Passing to Payload')

        # enter a ieee_addresses context before decoding the payload so that the current
        # hw source & destination addresses are known to the upper layers
        # (needed by 6lowpan-hc)
        with SixLowpanIPHC.encapsulating_iid_context (values["src"], values["dst"]):
            #log.debug('Starting to decode 802.15.4 payload as 6lowpan')
            # Payload
            #bin_slice = decode_field ("pl", bin_slice, SixLowpan)
            #TODO: reimplement the fall-back in a smart way (which is passive-friendly)

            try:
                bin_slice = decode_field ("pl", bin_slice, SixLowpan)
            except IndexError as e:
                # TODO: report this in a smarter way
                log.warning("Warning: unable to decode IEEE 802.15.4 payload as SixLowpan, ma "
                            "entering fallback mechanism")
                bin_slice = decode_field ("pl", bin_slice[:-2],SixLowpan)

        # let's make an educated guess and assume there's FCS if left bin_slice == 2
        if len(bin_slice) == 2:
            bin_slice = decode_FCS(bin_slice)

        return cls (*values), bin_slice

# class Ieee802154Ack (
#     metaclass = PacketClass,
#     variant_of = Ieee802154,
#     prune = 2, # I want to just keep FrameType, I'll add again seq and FCS as not consecutives fileds in parent class
#     fields = [
#          ("SequenceNumber", "seq", UInt8, 0),
#         ('FCS', 'fcs', Optional(HexUInt16), _Reversed(Omit()))
#     ]):
#     pass


import  ttproto.core.lib.ethernet
ttproto.core.lib.ethernet.ethernet_type_bidict[0x809a] = Ieee802154



import os
import pika
import logging
import threading
import traceback

# for using it as library and as a __main__
try:
    from messages import *
except:
    from .messages import *

VERSION = '0.0.10'
AMQP_EXCHANGE = 'amq.topic'
MAX_LOG_LINE_LENGTH = 120

logger = logging.getLogger(__name__)

class AmqpSynchCallTimeoutError(Exception):
    pass


class AmqpListener(threading.Thread):
    DEFAULT_TOPIC_SUSBCRIPTIONS = ['#']
    DEFAULT_EXCHAGE = 'amq.topic'
    DEFAULT_AMQP_URL = 'amqp://guest:guest@locahost/'

    def __init__(self, amqp_url, amqp_exchange, callback, topics=None, use_message_typing=True, pre_declared_queue=None):

        self.COMPONENT_ID = 'amqp_listener_%s' % str(uuid.uuid4())[:8]

        self.stopping = False
        self.connection = None
        self.channel = None
        if pre_declared_queue:
            self.services_queue_name = pre_declared_queue
            self.queue_is_declared = True
        else:
            self.services_queue_name = 'services_queue@%s' % self.COMPONENT_ID
            self.queue_is_declared = False
        self.use_message_typing = use_message_typing

        threading.Thread.__init__(self)

        if callback is None:
            self.message_dispatcher = AmqpListener.default_message_handler
        else:
            self.message_dispatcher = callback

        if topics:  # subscribe only to passed list
            self.topics = topics
        else:
            self.topics = self.DEFAULT_TOPIC_SUSBCRIPTIONS

        if amqp_exchange:
            self.exchange = amqp_exchange
        else:
            self.exchange = self.DEFAULT_EXCHAGE

        if amqp_url:
            self.amqp_url = amqp_url
        else:
            self.amqp_url = self.DEFAULT_AMQP_URL

        self.amqp_connect()

    @classmethod
    def default_message_handler(cls, message_as_dict):
        clean_dict = dict((k, v) for k, v in message_as_dict.items() if v)
        print('-' * 120)
        print('%s : %s' % ('routing_key', clean_dict.pop('routing_key')))
        print('-' * 120)
        print(json.dumps(clean_dict, indent=4, sort_keys=True))

    def amqp_connect(self):
        self.connection = pika.BlockingConnection(pika.URLParameters(self.amqp_url))
        self.channel = self.connection.channel()

        # Declare queue if necessary
        if not self.queue_is_declared:
            self.channel.queue_declare(queue=self.services_queue_name,
                                       auto_delete=True,
                                       arguments={'x-max-length': 200})
            for t in self.topics:
                self.channel.queue_bind(exchange=self.exchange,
                                        queue=self.services_queue_name,
                                        routing_key=t)
        # Hello world message
        m = MsgTestingToolComponentReady(
            component=self.COMPONENT_ID,
            description="%s is READY" % self.COMPONENT_ID

        )

        # Send hello world message
        self.channel.basic_publish(
            body=m.to_json(),
            routing_key=m.routing_key,
            exchange=self.exchange,
            properties=pika.BasicProperties(
                content_type='application/json',
            )
        )

        # Start consuming
        self.channel.basic_qos(prefetch_count=1)
        self.channel.basic_consume(self.on_request, queue=self.services_queue_name)

    def stop(self):
        self.stopping = True
        self.channel.queue_delete(self.services_queue_name)
        self.channel.stop_consuming()
        self.connection.close()

    def on_request(self, ch, method, props, body):

        if self.use_message_typing:
            try:
                m = Message.load_from_pika(method, props, body)
                if m is None:
                    raise Exception("Couldnt build message from json %s, rkey: %s " % (body, method.routing_key))
                m.routing_key = method.routing_key
                logger.debug('Message in bus: %s' % repr(m)[:MAX_LOG_LINE_LENGTH])
                self.message_dispatcher(m)

            except NonCompliantMessageFormatError as e:
                logger.error('%s got a non compliant message error %s' % (self.__class__.__name__, e))

            except Exception as e:
                logger.error(e)
                logger.error('message received:\n\tr_key: %s\n\t%s' % (method.routing_key, body))
                raise e

            finally:
                ch.basic_ack(delivery_tag=method.delivery_tag)
        else:

            props_dict = {
                'content_type': props.content_type,
                'delivery_mode': props.delivery_mode,
                'correlation_id': props.correlation_id,
                'reply_to': props.reply_to,
                'message_id': props.message_id,
                'timestamp': props.timestamp,
                'user_id': props.user_id,
                'app_id': props.app_id,
            }

            body_dict = json.loads(body.decode('utf-8'), object_pairs_hook=OrderedDict)
            ch.basic_ack(delivery_tag=method.delivery_tag)
            text_based_message_representation = OrderedDict()
            text_based_message_representation.update({'routing_key': method.routing_key})
            text_based_message_representation.update(props_dict)
            text_based_message_representation.update(body_dict)
            self.message_dispatcher(text_based_message_representation)

    def run(self):
        logger.info("Starting thread listening on the event bus on topics %s" % self.topics)

        for i in range(1, 4):
            try:
                self.channel.start_consuming()

            except (pika.exceptions.ConnectionClosed,
                    pika.exceptions.ChannelError,
                    pika.exceptions.ChannelClosed) as err:
                if not self.stopping:
                    logger.error('[AmqpListener] Unexpected connection closed, reconnecting %s/%s' % (i, 4))
                    logger.error(traceback.format_exc())
                    self.amqp_connect()

        logger.info('%s says Bye byes!' % self.COMPONENT_ID)


def publish_message(connection, message):
    """
    Publishes message into the correct topic (uses Message object metadata)
    Creates temporary channel on it's own
    Connection must be a pika.BlockingConnection
    """

    channel = connection.channel()
    properties = pika.BasicProperties(**message.get_properties())
    channel.basic_publish(
        exchange=AMQP_EXCHANGE,
        routing_key=message.routing_key,
        properties=properties,
        body=message.to_json(),
    )

    if channel and channel.is_open:
        channel.close()


def amqp_request(connection, request_message, component_id, retries=10, time_between_retries=0.5, use_message_typing=False):
    """
    Publishes message into the correct topic (uses Message object metadata)
    Returns reply message.
    Uses reply_to and corr id amqp's properties for matching the reply
    Creates temporary channel, and queues on it's own
    Connection must be a pika.BlockingConnection
    """

    # check first that sender didnt forget about reply to and corr id
    assert request_message.reply_to
    assert request_message.correlation_id
    assert retries > 0

    channel = None
    response = None

    reply_queue_name = 'amqp_rpc_%s@%s' % (str(uuid.uuid4())[:8], component_id)

    channel = connection.channel()
    result = channel.queue_declare(queue=reply_queue_name, auto_delete=True)
    callback_queue = result.method.queue

    # bind and listen to reply_to topic
    channel.queue_bind(
        exchange=AMQP_EXCHANGE,
        queue=callback_queue,
        routing_key=request_message.reply_to
    )

    channel.basic_publish(
        exchange=AMQP_EXCHANGE,
        routing_key=request_message.routing_key,
        properties=pika.BasicProperties(**request_message.get_properties()),
        body=request_message.to_json(),
    )

    time.sleep(0.2)
    retries_left = retries

    while retries_left > 0:
        time.sleep(time_between_retries)
        method, props, body = channel.basic_get(reply_queue_name)
        if method:
            channel.basic_ack(method.delivery_tag)
            if hasattr(props, 'correlation_id') and props.correlation_id == request_message.correlation_id:
                break
        retries_left -= 1

    if retries_left > 0:

        if use_message_typing:
            try:
                response = Message.load_from_pika(method, props, body)
                if response is None:
                    raise Exception("Couldnt build message from json %s, rkey: %s " % (body, method.routing_key))

            except NonCompliantMessageFormatError as e:
                logger.error('amqp_request got a non compliant message error %s' %  e)

            except Exception as e:
                logger.error(e)
                logger.error('message received:\n\tr_key: %s\n\t%s' % (method.routing_key, body))
                raise e

        else:

            body_dict = json.loads(body.decode('utf-8'), object_pairs_hook=OrderedDict)
            response = MsgReply(request_message, **body_dict)
    else:
        # clean up
        channel.queue_delete(reply_queue_name)
        channel.close()
        raise AmqpSynchCallTimeoutError(
            "Response timeout! rkey: %s , request type: %s" % (
                request_message.routing_key,
                type(request_message)
            )
        )

    if channel and channel.is_open:
        # clean up
        channel.queue_delete(reply_queue_name)
        channel.close()

    return response

if __name__ == '__main__':

    try:
        AMQP_EXCHANGE = str(os.environ['AMQP_EXCHANGE'])
    except KeyError as e:
        AMQP_EXCHANGE = "amq.topic"

    try:
        AMQP_URL = str(os.environ['AMQP_URL'])
        print('Env vars for AMQP connection succesfully imported')

    except KeyError as e:
        AMQP_URL = "amqp://guest:guest@localhost/"

    def callback_function(message_received):
        print("Callback function received: \n\t" + repr(message_received))


    # EXAMPLE ON AMQP LISTENER

    # # amqp listener example:
    # amqp_listener_thread = AmqpListener(
    #     amqp_url=AMQP_URL,
    #     amqp_exchange=AMQP_EXCHANGE,
    #     callback=callback_function,
    #     topics='#'
    # )
    #
    # try:
    #     amqp_listener_thread.start()
    # except Exception as e:
    #     print(e)
    #
    # # publish message example
    # retries_left = 3
    # while retries_left > 0:
    #     try:
    #         connection = pika.BlockingConnection(pika.URLParameters(AMQP_URL))
    #         m = MsgTest()
    #         publish_message(connection, m)
    #         break
    #     except pika.exceptions.ConnectionClosed:
    #         retries_left -= 1
    #         print('retrying..')
    #         time.sleep(0.2)
    #
    # # example of a request sent into the bus
    # m = MsgTestSuiteGetTestCases()
    # try:
    #     r = amqp_request(connection, m, 'someImaginaryComponent')
    #     print("This is the response I got:\n\t" + repr(r))
    #
    # except AmqpSynchCallTimeoutError as e:
    #     print("Nobody answered to our request :'(")


    # EXAMPLE ON REQUEST REPLY FOR UI BUTTONS:
    con = pika.BlockingConnection(pika.URLParameters(AMQP_URL))
    channel = con.channel()

    # ui_request = MsgUiRequestConfirmationButton()
    # print("publishing .. %s" % repr(ui_request))
    # ui_reply = amqp_request(con, ui_request, 'dummy_component')
    # print(repr(ui_reply))


    req = MsgUiRequestSessionConfiguration()
    ui_reply = amqp_request(con, req,'dummy_component')
    print(ui_reply.users)


from ..common import *


class TD_LWM2M_1_INT_301(CoAPTestCase):
    """
---
testcase_id: TD_LWM2M_1.0_INT_301
uri : http://openmobilealliance.org/iot/lightweight-m2m-lwm2m
configuration: LWM2M_CFG_01
objective: 
  - Sending the observation policy to the device
pre_conditions: 
  - The Client supports the Configuration C.4
  - Device is switched on and operational.
  - Client is registred with the LwM2M Server
notes: Please refer to TD_LWM2M_PRO.yaml file for more details
references: 'OMA-ETS-LightweightM2M-V1_0_1-20170926-A'

    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [CoAP(type='con', code='get'), CoAP(type='con', code='put')]

    def run(self):
        self.match('server', CoAP(type='con', code='get', opt=self.uri('/3/0/6')), 'fail')
        
        self.next()

        self.match('client', CoAP(code=2.05, pl=Not(b'')), 'fail')

        self.next()

        self.match('server', CoAP(type='con', code='put', opt=self.uri('/3/0/7')), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionUriQuery('pmin=5'))), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionUriQuery('pmax=15'))), 'fail')

        self.next()

        self.match('client', CoAP(code=Any(65, 68)), 'fail')

        self.next()

        self.match('server', CoAP(type='con', code='put', opt=self.uri('/3/0/8')), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionUriQuery('pmin=10'))), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionUriQuery('pmax=20'))), 'fail')

        self.next()

        self.match('client', CoAP(code=Any(65, 68)), 'fail')

        self.next()

        self.match('server', CoAP(type='con', code='get', opt=self.uri('/3/0/7')), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionObserve('0'))))

        self.next()

        self.match('client', CoAP(code=2.05), 'fail')

        self.next()

        self.match('server', CoAP(type='con', code='get', opt=self.uri('/3/0/8')), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionObserve('0'))))

        self.next()

        self.match('client', CoAP(code=2.05), 'fail')




from ..common import *

class TD_COAP_OBS_01(CoAPTestCase):
    """
---
TD_COAP_OBS_01:
    cfg: CoAP_CFG_BASIC
    obj: Handle resource observation with CON messages
    ref: '[OBSERVE] 1.2, 3, 4'
    pre:
        - Client supports Observe option
        - Server supports Observe option
        - Server offers an observable resource /obs which changes periodically (e.g. every 5s) which produces confirmable notifications
    seq:
    -   s: Client is requested to send to the server a confirmable GET
           request with observe option for resource /obs

    -   c:
        - 'The request sent by client contains:'
        -   - Type = 0(CON)
            - Code = 1(GET)
            - Token value = a value generated by the client
            - Observe option = empty
    -   c:
        - 'Server sends the response containing:'
        -   - Type = 2(ACK)
            - Code = 2.05(Content)
            - Content-format of the resource /obs
            - Token value = same as one found in the step 2
            - Observe option with a sequence number

    -   c:
        - 'Server sends a notification containing:'
        -   - Type = 0(CON)
            - Code = 2.05(Content)
            - Content-format =  same as one found in the step 3
            - Token value = same as one found in the step 3
            - "Observe option indicating increasing values(sequence number arithmetic modulo 2**24), unless more than 256 seconds elapsed"

    -   c:
        --   s: Client is requested to send to the server a confirmable GET
           request with observe option for resource /obs

    -   c:
        - 'The request sent by client contains:'
        -   - Type = 0(CON)
            - Code = 1(GET)
            - Token value = a value generated by the client
            - Observe option = empty
    -   c:
        - 'Server sends the response containing:'
        -   - Type = 2(ACK)
            - Code = 2.05(Content)
            - Content-format of the resource /obs
            - Token value = same as one found in the step 2
            - Observe option with a sequence number

    -   c:
        - 'Server sends a notification containing:'
        -   - Type = 0(CON)
            - Code = 2.05(Content)
            - Content-format =  same as one found in the step 3
            - Token value = same as one found in the step 3
            - "Observe option indicating increasing values(sequence number arithmetic modulo 2**24), unless more than 256 seconds elapsed"

    -   c:
        - Client sends an ACK
    -   v: Client displays the received information Client sends an ACK
    -   v: Client displays the received information
    """
    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [
            CoAP(type='con', code='get', opt=Opt(CoAPOptionObserve(0), CoAPOptionUriPath("obs")))
        ]

    def run(self):
        # Step 2
        self.match("client", CoAP(type="con", code="get",
        opt=self.uri("/obs", CoAPOptionObserve(0))))

        token = self.coap["tok"]

        self.next()

        # Step 3
        if not self.match("server", CoAP(type="ack", code=2.05, opt=Opt(
            CoAPOptionObserve(),
            CoAPOptionContentFormat(),
        ))):
                raise self.Stop()
                self.match("server", CoAP(tok=token), "fail")

        content_format = self.coap["opt"][CoAPOptionContentFormat]["val"]
        index          = self.coap["opt"][CoAPOptionObserve]["val"]

        # match at least 2 notifications
        for i in range(2):

            self.next()

            # Step 4
            if not self.match("server", CoAP(type="con", code=2.05, opt=Opt(
                CoAPOptionObserve(),
                CoAPOptionContentFormat(),
            ))):
                raise self.Stop()
                self.match("server", CoAP(tok=token), "fail")
                self.match("server", CoAP(opt=Opt(CoAPOptionContentFormat(content_format))), "fail")

            new_index = self.coap["opt"][CoAPOptionObserve]["val"]

            self.set_verdict(("pass" if new_index > index else "fail"),
                             "value of observe option must be increasing")

            self.next()

            # Step 6
            self.match("client", CoAP(type="ack", code=0))

        raise self.Stop()



import unittest

from ttproto.core.analyzer import Verdict
from ttproto.core.typecheck3000 import InputParameterError


class VerdictTestCase(unittest.TestCase):
    """
    Test class for the verdict class
    """

    # #################### Init and deinit functions #########################
    def setUp(self):
        """
            Initialize the verdict instance
        """
        self.verdict = Verdict()

    # #################### Tests functions #########################

    # ##### __init__
    def test___init__(self):
        self.assertEqual(self.verdict.get_value(), 'none')
        self.assertEqual(self.verdict.get_message(), '')
        self.assertEqual(str(self.verdict), 'none')

    def test___init__with_initial_value(self):
        for verdict in Verdict.values():
            self.verdict = Verdict(verdict)
            self.assertEqual(
                self.verdict.get_value(),
                verdict
            )
            self.assertEqual(self.verdict.get_message(), '')
            self.assertEqual(
                str(self.verdict),
                verdict
            )

    def test___init__with_int_as_initial_value(self):
        verdict_values = Verdict.values()
        for verdict in verdict_values:
            verdict_id = verdict_values.index(verdict)
            with self.assertRaises(InputParameterError):
                self.verdict = Verdict(verdict_id)

    # ##### update
    def test_update(self):
        for verdict in Verdict.values():
            if verdict != 'none':  # Because initial one is none
                self.verdict.update(verdict, verdict)
                self.assertEqual(self.verdict.get_value(), verdict)
                self.assertEqual(self.verdict.get_message(), verdict)
                self.assertEqual(str(self.verdict), verdict)

    def test_update_with_lower_or_equal_values(self):
        verdict_values = Verdict.values()
        for verdict in verdict_values:

            # Initialize with the verdict
            self.verdict = Verdict(verdict)

            # Try to update with lower or equal values
            for ver in verdict_values:
                if verdict_values.index(ver) <= verdict_values.index(verdict):
                    self.verdict.update(ver, ver)
                    self.assertEqual(self.verdict.get_value(), verdict)
                    self.assertEqual(self.verdict.get_message(), '')
                    self.assertEqual(str(self.verdict), verdict)

    def test_update_with_int_value(self):
        verdict_values = Verdict.values()
        for verdict in verdict_values:

            # Initialize with the verdict
            self.verdict = Verdict(verdict)

            # Try to update with lower or equal values
            for ver in verdict_values:
                next_verdict_id = verdict_values.index(ver)
                with self.assertRaises(InputParameterError):
                    self.verdict.update(next_verdict_id, '')

    # ##### values
    def test_values(self):
        verdict_values = Verdict.values()
        self.assertIsInstance(verdict_values, tuple)
        self.assertGreater(len(verdict_values), 0)
        for verdict in verdict_values:
            self.assertIsInstance(verdict, str)

    # ##### get_value
    def test_get_value(self):
        for verdict in Verdict.values():

            # Initialize with the verdict
            self.verdict = Verdict(verdict)

            # Try to get its value
            self.assertEqual(self.verdict.get_value(), verdict)

    # ##### get_message
    def test_get_message(self):
        verdict_values = Verdict.values()
        for verdict in Verdict.values():

            # Initialize with the verdict
            self.verdict = Verdict(verdict)
            next_verdict_id = verdict_values.index(verdict) + 1

            if next_verdict_id < len(verdict_values):
                self.verdict.update(verdict_values[next_verdict_id], verdict)
                self.assertEqual(self.verdict.get_message(), verdict)

    # ##### __str__
    def test___str__(self):
        for verdict in Verdict.values():

            # Initialize with the verdict
            self.verdict = Verdict(verdict)

            # Try to get its value
            self.assertEqual(str(self.verdict), verdict)

# #################### Main run the tests #########################
if __name__ == '__main__':
    unittest.main()



import unittest

from collections import OrderedDict
from ttproto.core.dissector import Capture, Dissector, Frame, ReaderError
from ttproto.core.lib.all import *
from tests.test_tools.struct_validator import StructureValidator
from ttproto.core.packet import Value, PacketValue
from ttproto.core.lib.inet.meta import InetPacketValue
from ttproto.core.typecheck3000 import InputParameterError


class FrameTestCase(unittest.TestCase):
    """
    Test class for the Frame object
    """

    # #################### Tests parameters #########################

    # File path
    TEST_FILE_DIR = 'tests/test_dumps'
    PCAP_FILE = TEST_FILE_DIR + '/coap/CoAP_plus_random_UDP_messages.pcap'

    # Create a struct checker object
    struct_validator = StructureValidator()
    FRAMES_PROTOCOL = {
        1: [
            UDP,
            IPv4,
            NullLoopback
        ],
        2: [
            UDP,
            IPv4,
            NullLoopback
        ],
        3: [
            IPv4,
            NullLoopback
        ],
        4: [
            UDP,
            IPv4,
            NullLoopback,
            CoAP
        ],
        5: [
            UDP,
            IPv4,
            NullLoopback,
            CoAP
        ]
    }
    FRAME_VALUES = {
        'id': int,
        'ts': float,
        'error': Exception,
        'value': Value
    }

    # #################### Utility functions #########################

    def list_diff(self, list1, list2):
        return_list = []
        for el in list2:
            if el not in list1:
                return_list.append(el)
        return return_list

    def frames_with_protocol(self, protocol):
        frames_with_this_protocol = []
        for frame_prot in self.FRAMES_PROTOCOL:
            if protocol in self.FRAMES_PROTOCOL[frame_prot]:
                frames_with_this_protocol.append(frame_prot)
        return frames_with_this_protocol

    # #################### Init and deinit functions #########################
    def setUp(self):
        """
            Initialize the frame list from the valid pcap file
        """
        self.frames = Capture(self.PCAP_FILE).frames

    # #################### Tests functions #########################

    # ##### __repr__
    def test___repr__(self):

        # Check the str representation
        for frame in self.frames:
            self.assertEqual(type(str(frame)), str)
            self.assertEqual(type(repr(frame)), str)
            self.assertGreater(len(str(frame)), 0)
            self.assertGreater(len(repr(frame)), 0)

    # ##### summary
    def test_summary(self):

        # Check the str representation
        counter = 1
        for frame in self.frames:

            # Get the summary
            summary = frame.summary()

            # Check summary structure
            self.assertIsNotNone(summary)
            self.assertEqual(type(summary), tuple)
            self.assertEqual(len(summary), 2)

            # Check first element
            self.assertEqual(type(summary[0]), int)
            self.assertEqual(summary[0], counter)

            # Check second element
            self.assertEqual(type(summary[1]), str)
            self.assertGreater(len(summary[1]), 0)

            # Increment counter
            counter += 1

    # ##### dict
    def test_dict(self):

        # Check the str representation
        for frame in self.frames:

            # Get the dictionnary
            dictionnary = frame.dict()

            # Check dictionnary structure
            self.assertIsNotNone(dictionnary)
            self.assertEqual(type(dictionnary), OrderedDict)
            self.assertEqual(len(dictionnary), 5)

            # Check whole structure from struct checker
            self.struct_validator.check_frame(dictionnary)

    # ##### __contains__
    def test___contains__(self):

        # Check the provided structure
        self.assertEqual(len(self.FRAMES_PROTOCOL), len(self.frames))

        # Get all the protocols
        protocols = Dissector.get_implemented_protocols()

        # Check that the protocols are in the correct frame
        i = 1
        for frame in self.frames:

            # Get the elements that shouldn't be in it
            should_not_be = self.list_diff(self.FRAMES_PROTOCOL[i], protocols)

            # Check that those which shouldn't be present really aren't
            for non_present in should_not_be:
                self.assertNotIn(non_present, frame)

            # Check that those which are present really are
            for prot in self.FRAMES_PROTOCOL[i]:
                self.assertIn(prot, frame)

            # Increment counter
            i += 1

    def test___contains__none(self):

        # Check that the protocols are in the correct frame
        for frame in self.frames:

            # Expect the exception
            with self.assertRaises(InputParameterError):
                check = None in frame

    def test___contains__not_a_protocol(self):

        # Check that the protocols are in the correct frame
        for frame in self.frames:

            # Expect the exception
            with self.assertRaises(InputParameterError):
                check = Frame in frame

    def test___contains__higher_classes(self):

        # Check that the protocols are in the correct frame
        for frame in self.frames:

            # Expect the exception
            with self.assertRaises(InputParameterError):
                check = InetPacketValue in frame

            # Expect the exception
            with self.assertRaises(InputParameterError):
                check = PacketValue in frame

            # Expect the exception
            with self.assertRaises(InputParameterError):
                check = Value in frame

    # ##### filter_frames
    def test_filter_frames(self):

        # Get all the protocols
        protocols = Dissector.get_implemented_protocols()

        # Filter on each protocol
        for protocol in protocols:

            # Filter on each protocol
            filtered, ignored = Frame.filter_frames(self.frames, protocol)

            # Get the id of frames with this protocol
            ids = self.frames_with_protocol(protocol)

            # Check the two datas received
            self.assertEqual(type(filtered), list)
            for f in filtered:
                self.assertEqual(type(f), Frame)
            self.assertEqual(type(ignored), list)
            for i in ignored:
                self.assertEqual(type(i), Frame)
            self.assertEqual(len(filtered) + len(ignored), len(self.frames))

            # Check the length of filtered
            self.assertEqual(len(filtered), len(ids))

            # Check that each element goes together
            for frame in filtered:
                dictionnary = frame.dict()
                self.assertIn(dictionnary['id'], ids)

    def test_filter_frames_none_type(self):

        # Filter on none protocol
        with self.assertRaises(InputParameterError):
            filtered, ignored = Frame.filter_frames(self.frames, type(None))

    def test_filter_frames_none(self):

        # Filter on none protocol
        with self.assertRaises(InputParameterError):
            filtered, ignored = Frame.filter_frames(self.frames, None)

    def test_filter_frames_not_a_protocol(self):

        # Filter on each protocol
        with self.assertRaises(InputParameterError):
            filtered, ignored = Frame.filter_frames(self.frames, Frame)

    def test_filter_frames_higher_classes(self):

        # Filter on each protocol
        with self.assertRaises(InputParameterError):
            filtered, ignored = Frame.filter_frames(
                self.frames,
                InetPacketValue
            )
        with self.assertRaises(InputParameterError):
            filtered, ignored = Frame.filter_frames(self.frames, PacketValue)
        with self.assertRaises(InputParameterError):
            filtered, ignored = Frame.filter_frames(self.frames, Value)

    def test_filter_frames_only_single_frame(self):

        # Get all the protocols
        protocols = Dissector.get_implemented_protocols()

        # Filter on each protocol
        for protocol in protocols:

            # Filter on none protocol
            with self.assertRaises(InputParameterError):
                filtered, ignored = Frame.filter_frames(
                    self.frames[0],
                    protocol
                )

    def test_filter_frames_list_of_non_frame(self):

        # Get all the protocols
        protocols = Dissector.get_implemented_protocols()

        # Filter on each protocol
        for protocol in protocols:

            # Filter on none protocol
            with self.assertRaises(TypeError):
                filtered, ignored = Frame.filter_frames(protocols, protocol)

    def test_filter_frames_list_of_frame_with_a_non_frame(self):

        # Get all the protocols
        protocols = Dissector.get_implemented_protocols()

        # Insert a non frame object
        self.frames.insert(1, protocols[1])

        # Filter on each protocol
        for protocol in protocols:

            # Filter on none protocol
            with self.assertRaises(TypeError):
                filtered, ignored = Frame.filter_frames(self.frames, protocol)

    # ##### __getitem__
    def test___getitem__values(self):

        # For each value name, check its type
        for value_name, value_type in self.FRAME_VALUES.items():

            if value_name != 'error':
                for frame in self.frames:
                    self.assertIsInstance(frame[value_name], value_type)
            else:  # 'error' field can be None
                for frame in self.frames:
                    if frame[value_name]:
                        self.assertIsInstance(frame[value_name], value_type)

    def test___getitem__unknown_value(self):

        for frame in self.frames:
            with self.assertRaises(AttributeError):
                test = frame['unknown']

    def test___getitem__none(self):
        for frame in self.frames:
            with self.assertRaises(InputParameterError):
                test = frame[None]

    def test___getitem__type_but_not_a_protocol(self):

        # Check that the protocols are in the correct frame
        for frame in self.frames:

            # Expect the exception
            with self.assertRaises(InputParameterError):
                test = frame[Frame]

    def test___getitem__higher_classes(self):

        # Check that the protocols are in the correct frame
        for frame in self.frames:

            # Expect the exception
            with self.assertRaises(InputParameterError):
                check = frame[InetPacketValue]

            # Expect the exception
            with self.assertRaises(InputParameterError):
                check = frame[PacketValue]

            # Expect the exception
            with self.assertRaises(InputParameterError):
                check = frame[Value]


# #################### Main run the tests #########################
if __name__ == '__main__':
    unittest.main()
